#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-11-12
# @Filename: cli.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import logging
import os
import signal
import socket
import sys
import warnings
from copy import deepcopy
from functools import wraps

from typing import Optional

import click
from click_default_group import DefaultGroup

from sdsstools.daemonizer import DaemonGroup

from jaeger import can_log, config, log
from jaeger.commands.bootloader import load_firmware
from jaeger.commands.calibration import calibrate_positioners
from jaeger.commands.goto import goto as goto_
from jaeger.exceptions import (
    FPSLockedError,
    JaegerError,
    JaegerUserWarning,
    TrajectoryError,
)
from jaeger.fps import FPS, LOCK_FILE
from jaeger.positioner import Positioner
from jaeger.testing import VirtualFPS


__FPS__ = None


def shutdown(sign):
    """Shuts down the FPS and stops the positioners in case of a signal interrupt."""

    if __FPS__ is not None:
        __FPS__.send_command("SEND_TRAJECTORY_ABORT", positioner_ids=None, now=True)
        log.error(f"stopping positioners and cancelling due to {sign.name}")
        sys.exit(0)
    else:
        log.error(f"cannot shutdown FPS before {sign.name}")
        sys.exit(1)


def cli_coro(f):
    """Decorator function that allows defining coroutines with click."""

    @wraps(f)
    def wrapper(*args, **kwargs):
        loop = asyncio.get_event_loop()
        signals = (signal.SIGHUP, signal.SIGTERM, signal.SIGINT)
        for ss in signals:
            loop.add_signal_handler(ss, shutdown, ss)
        return loop.run_until_complete(f(*args, **kwargs))

    return wrapper


class FPSWrapper(object):
    """A helper to store FPS initialisation parameters."""

    def __init__(
        self,
        profile,
        ieb=None,
        initialise=True,
        npositioners=10,
        enable_low_temperature=True,
        skip_fibre_assignments_check=False,
    ):

        self.profile = profile
        if self.profile in ["test", "virtual"]:
            self.profile = "virtual"

        self.ieb = ieb
        self.enable_low_temperature = enable_low_temperature
        self.skip_fibre_assignments_check = skip_fibre_assignments_check
        self.initialise = initialise

        self.vpositioners = []
        self.npositioners = npositioners

        self.fps = None

    async def __aenter__(self):

        global __FPS__

        # If profile is test we start a VirtualFPS first so that it can respond
        # to the FPS class.
        if self.profile == "virtual":
            self.fps = VirtualFPS()
            for pid in range(self.npositioners):
                self.fps.add_virtual_positioner(pid + 1)
        else:
            self.fps = FPS(can=self.profile, ieb=self.ieb)

        __FPS__ = self.fps

        if self.initialise:
            await self.fps.initialise(
                enable_low_temperature=self.enable_low_temperature,
                skip_fibre_assignments_check=self.skip_fibre_assignments_check,
            )
            if self.enable_low_temperature is False:
                warnings.warn(
                    "Disabling low temperature handling for sextant.",
                    JaegerUserWarning,
                )

        return self.fps

    async def __aexit__(self, *excinfo):
        try:
            if self.fps is None:
                return
            await self.fps.shutdown()
        except JaegerError as err:
            warnings.warn(f"Failed shutting down FPS: {err}", JaegerUserWarning)


pass_fps = click.make_pass_decorator(FPSWrapper, ensure=True)


@click.group(cls=DefaultGroup, default="actor", default_if_no_args=True)
@click.option(
    "-c",
    "--config",
    "config_file",
    type=click.Path(exists=True, dir_okay=False),
    help="Path to the user configuration file.",
)
@click.option(
    "-p",
    "--profile",
    type=str,
    help="The bus interface profile.",
)
@click.option(
    "--virtual",
    is_flag=True,
    help="Runs a virtual FPS with virtual positioners. Same as --profile=virtual.",
)
@click.option(
    "-n",
    "--npositioners",
    type=int,
    default=10,
    help="How many virtual positioners must be connected to the virtual FPS.",
)
@click.option(
    "-v",
    "--verbose",
    count=True,
    help="Debug mode. Use additional v for more details.",
)
@click.option(
    "-q",
    "--quiet",
    is_flag=True,
    help="Disable all console logging.",
)
@click.option(
    "--ieb/--no-ieb",
    default=None,
    help="Does not connect to the IEB.",
)
@click.option(
    "-s",
    "--sextant",
    is_flag=True,
    help="Use engineering sextant instead of IEB. "
    "Modifies the internal configuration file.",
)
@click.option(
    "--allow-host",
    is_flag=True,
    help="Allows running jaeger in a host other than sdss5-fps.",
)
@click.option(
    "--no-lock",
    is_flag=True,
    help="Do not use the lock file, or ignore it if present.",
)
@click.option(
    "-x",
    "--skip-fibre-assignments-check",
    is_flag=True,
    help="Do not fail if the fibre assignment check fails.",
)
@click.pass_context
def jaeger(
    ctx,
    config_file,
    profile,
    verbose,
    quiet,
    ieb,
    sextant,
    virtual,
    npositioners,
    allow_host,
    no_lock,
    skip_fibre_assignments_check,
):
    """CLI for the SDSS-V focal plane system.

    If called without subcommand starts the actor.

    """

    if allow_host is False:
        hostname = socket.getfqdn()
        if hostname.endswith("apo.nmsu.edu") or hostname.endswith("lco.cl"):
            if not hostname.startswith("sdss5-fps"):
                raise RuntimeError(
                    "At the observatories jaeger must run on sdss5-fps. "
                    "If you want to run jaeger on another computer use --allow-host."
                )

    if verbose > 0 and quiet:
        raise click.UsageError("--quiet and --verbose are mutually exclusive.")

    if config_file:
        config.load(config_file)

    if no_lock:
        config["fps"]["use_lock"] = False

    actor_config = config.get("actor", {})

    if verbose == 1:
        log.sh.setLevel(logging.INFO)
        actor_config["verbose"] = logging.INFO
    elif verbose == 2:
        log.sh.setLevel(logging.DEBUG)
        actor_config["verbose"] = logging.DEBUG
    elif verbose >= 3:
        log.sh.setLevel(logging.DEBUG)
        can_log.sh.setLevel(logging.DEBUG)
        actor_config["verbose"] = logging.DEBUG

    if quiet:
        log.handlers.remove(log.sh)
        warnings.simplefilter("ignore")
        actor_config["verbose"] = 100

    if sextant:
        sextant_file = os.path.join(os.path.dirname(__file__), "etc/sextant.yaml")
        config["ieb"]["config"] = sextant_file
        log.debug(f"Using internal IEB sextant onfiguration file {sextant_file}.")

    if sextant or "sextants/" in config["ieb"]["config"]:
        enable_low_temperature = False
    else:
        enable_low_temperature = True

    if virtual is True:
        profile = "virtual"

    ctx.obj = FPSWrapper(
        profile,
        ieb=ieb,
        npositioners=npositioners,
        enable_low_temperature=enable_low_temperature,
        skip_fibre_assignments_check=skip_fibre_assignments_check,
    )


LOG_FILE = os.path.join(
    os.environ.get("ACTOR_DAEMON_LOG_DIR", "$HOME/logs"),
    "jaeger/jaeger.log",
)


@jaeger.group(
    cls=DaemonGroup,
    prog="jaeger_actor",
    workdir=os.getcwd(),
    log_file=LOG_FILE,
)
@click.option(
    "--no-tron",
    is_flag=True,
    help="Does not connect to Tron.",
)
@pass_fps
@cli_coro
async def actor(fps_maker, no_tron: bool = False):
    """Runs the actor."""

    try:
        from jaeger.actor import JaegerActor
    except ImportError:
        raise ImportError("CLU needs to be installed to run jaeger as an actor.")

    config_copy = deepcopy(config)
    if "actor" not in config_copy:
        raise RuntimeError("Configuration file does not contain an actor section.")

    config_copy["actor"].pop("status", None)

    if no_tron:
        config_copy["actor"].pop("tron", None)

    # Do not initialise FPS so that we can define the actor instance first.
    fps_maker.initialise = False

    async with fps_maker as fps:
        actor_: JaegerActor = JaegerActor.from_config(config_copy, fps)

        await fps.initialise(
            skip_fibre_assignments_check=fps_maker.skip_fibre_assignments_check
        )

        await actor_.start()
        await actor_.run_forever()

        await actor_.stop()


@jaeger.command(name="upgrade-firmware")
@click.argument(
    "firmware-file",
    nargs=1,
    type=click.Path(exists=True),
)
@click.argument(
    "SEXTANTS",
    type=int,
    nargs=-1,
    required=False,
)
@click.option(
    "-y",
    "--yes",
    is_flag=True,
    help="Do not ask for confirmation.",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Forces skipping of invalid positioners",
)
@click.option(
    "-p",
    "--positioners",
    type=str,
    help="Comma-separated positioners to upgrade",
)
@click.option(
    "-n",
    "--no-cycle",
    is_flag=True,
    help="Does not power cycle positioners before upgrading each sextant",
)
@click.option(
    "-o",
    "--all-on",
    is_flag=True,
    help="Powers on all sextants after a successful upgrade.",
)
@pass_fps
@cli_coro
async def upgrade_firmware(
    fps_maker,
    firmware_file,
    force,
    positioners,
    no_cycle,
    sextants,
    all_on,
    yes,
):
    """Upgrades the firmaware."""

    if positioners is not None:
        positioners = [int(positioner.strip()) for positioner in positioners.split(",")]

    fps_maker.initialise = False

    sextants = sextants or (1, 2, 3, 4, 5, 6)

    if not yes:
        click.confirm(
            f"Upgrade firmware for sextant(s) {', '.join(map(str, sextants))}?",
            default=False,
            abort=True,
        )

    async with fps_maker as fps:

        ps_devs = []
        if fps.ieb and no_cycle is False:
            for module in fps.ieb.modules.values():
                for dev_name in module.devices:
                    if dev_name.upper().startswith("PS"):
                        ps_devs.append(dev_name)

        for sextant in sextants:
            log.info(f"Upgrading firmware on sextant {sextant}.")

            if fps.ieb and no_cycle is False:
                log.info("Power cycling positioners")

                for dev in ps_devs:
                    await fps.ieb.get_device(dev).open()
                    await asyncio.sleep(1)

                await asyncio.sleep(5)

                dev = "PS" + str(sextant)
                await fps.ieb.get_device(dev).close()

                await asyncio.sleep(3)

            await fps.initialise(start_pollers=False)

            await load_firmware(
                fps,
                firmware_file,
                positioners=positioners,
                force=force,
                show_progressbar=True,
            )

    if all_on is True:
        log.info("Powering on sextants.")
        for sextant in sextants:
            await fps.ieb.get_device(f"PS{sextant}").close()
            await asyncio.sleep(3)


@jaeger.command()
@click.argument("positioner-id", nargs=1, type=int)
@click.option(
    "--motors/--no-motors",
    is_flag=True,
    default=True,
    help="Run the motor calibration.",
)
@click.option(
    "--datums/--no-datums",
    is_flag=True,
    default=True,
    help="Run the datum calibration.",
)
@click.option(
    "--cogging/--no-cogging",
    is_flag=True,
    default=True,
    help="Run the cogging calibration (can take a long time).",
)
@pass_fps
@cli_coro
async def calibrate(fps_maker, positioner_id, motors, datums, cogging):
    """Runs a full calibration on a positioner."""

    fps_maker.initialise = False
    fps_maker.danger = True

    async with fps_maker as fps:
        await fps.initialise(start_pollers=False)
        await calibrate_positioners(
            fps,
            "both",
            positioner_id,
            motors=motors,
            datums=datums,
            cogging=cogging,
        )


@jaeger.command()
@click.argument("POSITIONER-IDS", type=int, nargs=-1)
@click.argument("ALPHA", type=click.FloatRange(-10.0, 370.0))
@click.argument("BETA", type=click.FloatRange(-10.0, 370.0))
@click.option(
    "-r",
    "--relative",
    is_flag=True,
    help="Whether this is a relative move",
)
@click.option(
    "-s",
    "--speed",
    type=click.FloatRange(100.0, 4000.0),
    help="The speed for both alpha and beta arms, in RPS on the input.",
)
@click.option(
    "-a",
    "--all",
    is_flag=True,
    default=False,
    help="Applies to all valid positioners.",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Forces a move to happen.",
)
@click.option(
    "--go-cowboy",
    is_flag=True,
    help="If set, does not use kaiju-validated trajectories.",
)
@click.option(
    "--use-sync/-no-use-sync",
    " /-S",
    default=True,
    help="Whether to use the SYNC line to start the trajectory.",
)
@pass_fps
@cli_coro
async def goto(
    fps_maker,
    positioner_ids: tuple[int, ...] | list[int],
    alpha: float,
    beta: float,
    speed: float | None,
    all: bool = False,
    force: bool = False,
    relative: bool = False,
    use_sync: bool = True,
    go_cowboy: bool = False,
):
    """Sends positioners to a given (alpha, beta) position."""

    with fps_maker as fps:

        if all:
            if not force:
                raise JaegerError("Use --force to move all positioners at once.")
            positioner_ids = list(fps.positioners.keys())
        else:
            positioner_ids = list(positioner_ids)

        if not relative:
            if alpha < 0 or beta < 0:
                raise JaegerError("Negative angles only allowed in relative mode.")

        new_positions = {}
        for pid in positioner_ids:
            new_positions[pid] = (alpha, beta)

        try:
            await goto_(
                fps,
                new_positions,
                speed=speed,
                relative=relative,
                use_sync_line=use_sync,
                go_cowboy=go_cowboy,
            )
        except (JaegerError, TrajectoryError) as err:
            raise JaegerError(f"Goto command failed: {err}")


@jaeger.command(name="set-positions")
@click.argument("positioner_id", metavar="POSITIONER", type=int)
@click.argument("alpha", metavar="ALPHA", type=float)
@click.argument("beta", metavar="BETA", type=float)
@pass_fps
@cli_coro
async def set_positions(fps_maker, positioner_id, alpha, beta):
    """Sets the position of the alpha and beta arms."""

    if alpha < 0 or alpha >= 360:
        raise click.UsageError("alpha must be in the range [0, 360)")

    if beta < 0 or beta >= 360:
        raise click.UsageError("beta must be in the range [0, 360)")

    async with fps_maker as fps:

        positioner = fps.positioners[positioner_id]

        result = await positioner.set_position(alpha, beta)

        if not result:
            log.error("failed to set positions.")
            return

        log.info(f"positioner {positioner_id} set to {(alpha, beta)}.")


@jaeger.command()
@click.argument("positioner_id", metavar="POSITIONER", type=int)
@click.option(
    "--axis",
    type=click.Choice(["alpha", "beta"], case_sensitive=True),
    help="The axis to home. If not set, homes both axes at the same time.",
)
@pass_fps
@cli_coro
async def home(fps_maker: FPSWrapper, positioner_id: int, axis: str | None = None):
    """Home a single positioner, sending a GO_TO_DATUMS command."""

    alpha: bool = axis == "alpha" or axis is None
    beta: bool = axis == "beta" or axis is None

    async with fps_maker as fps:

        if positioner_id not in fps or fps[positioner_id].initialised is False:
            raise ValueError("Positioner is not connected.")
        if fps[positioner_id].disabled or fps[positioner_id].offline:
            raise ValueError("Positioner has been disabled.")

        await fps[positioner_id].home(alpha=alpha, beta=beta)

    return


@jaeger.command()
@click.argument("positioner_id", metavar="POSITIONER", type=int)
@pass_fps
@cli_coro
async def status(fps_maker: FPSWrapper, positioner_id: int):
    """Returns the status of a positioner with low-level initialisation."""

    fps_maker.initialise = False

    async with fps_maker as fps:

        pos = Positioner(positioner_id, fps)

        try:
            await pos.update_firmware_version()
            print(f"Firmware: {pos.firmware}")
            print(f"Bootloader: {pos.is_bootloader()}")
        except Exception as err:
            raise JaegerError(f"Failed retrieving firmware: {err}")

        try:
            await pos.update_status()
            bit_names = ", ".join(bit.name for bit in pos.status.active_bits)
            print(f"Status: {pos.status.value} ({bit_names})")
        except Exception as err:
            raise JaegerError(f"Failed retrieving status: {err}")

        try:
            await pos.update_position()
            print(f"Position: {pos.position}")
        except Exception as err:
            raise JaegerError(f"Failed retrieving position: {err}")


@jaeger.command()
@click.option(
    "--collision-buffer",
    type=click.FloatRange(1.6, 3.0),
    help="Custom collision buffer",
)
@click.option(
    "--force",
    is_flag=True,
    help="Execute unwind even in presence of deadlocks.",
)
@pass_fps
@cli_coro
async def unwind(fps_maker, collision_buffer: float | None = None, force: bool = False):
    """Unwinds the array."""

    from jaeger.kaiju import unwind

    async with fps_maker as fps:

        if fps.locked:
            FPSLockedError("The FPS is locked.")

        await fps.update_position()
        positions = {p.positioner_id: (p.alpha, p.beta) for p in fps.values()}

        log.info("Calculating trajectory.")
        trajectory = await unwind(
            positions,
            collision_buffer=collision_buffer,
            disabled=[pid for pid in fps.positioners if fps.positioners[pid].disabled],
            force=force,
        )

        log.info("Executing unwind trajectory.")
        await fps.send_trajectory(trajectory)

    return


@jaeger.command()
@click.argument("EXPLODE-DEG", type=float)
@click.option("--one", type=int, help="Only explode this positioner.")
@pass_fps
@cli_coro
async def explode(
    fps_maker,
    explode_deg: float,
    one: int | None = None,
):
    """Explodes the array."""

    from jaeger.kaiju import explode

    async with fps_maker as fps:

        if fps.locked:
            FPSLockedError("The FPS is locked.")

        await fps.update_position()
        positions = {p.positioner_id: (p.alpha, p.beta) for p in fps.values()}

        log.info("Calculating trajectory.")
        trajectory = await explode(
            positions,
            explode_deg=explode_deg,
            disabled=[pid for pid in fps.positioners if fps.positioners[pid].disabled],
            positioner_id=one,
        )

        log.info("Executing explode trajectory.")
        await fps.send_trajectory(trajectory)

    return


@jaeger.command()
@click.argument("PATH", required=False, type=click.Path(exists=False, dir_okay=False))
@click.option("--collision-buffer", type=float, help="The collision buffer.")
@pass_fps
@cli_coro
async def snapshot(
    fps_maker: FPSWrapper,
    path: Optional[str] = None,
    collision_buffer: float | None = None,
):
    """Takes a snapshot image."""

    if path is not None:
        path = str(path)

    async with fps_maker as fps:

        await fps.update_position()
        filename = await fps.save_snapshot(path, collision_buffer=collision_buffer)

    print(f"Snapshot saved to {filename}")


@jaeger.command()
@pass_fps
@cli_coro
async def unlock(fps_maker: FPSWrapper):
    """Unlocks the FPS."""

    warnings.filterwarnings(
        "ignore",
        message=".+FPS was collided and has been locked.+",
        category=JaegerUserWarning,
    )

    async with fps_maker as fps:
        await fps.unlock()

    if os.path.exists(LOCK_FILE):
        os.remove(LOCK_FILE)


if __name__ == "__main__":
    jaeger()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-05-13
# @Filename: commands.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import json
import pathlib
from time import time

from typing import TYPE_CHECKING

import click
import numpy

import clu

from jaeger.can import JaegerCAN
from jaeger.commands import SetCurrent, Trajectory
from jaeger.commands.goto import goto as goto_
from jaeger.commands.trajectory import send_trajectory
from jaeger.exceptions import JaegerError, TrajectoryError

from . import JaegerCommandType, jaeger_parser


if TYPE_CHECKING:
    from clu import Command

    from jaeger import FPS
    from jaeger.actor import JaegerActor


__all__ = [
    "goto",
    "speed",
    "initialise",
    "stop",
    "hall",
    "trajectory",
    "unlock",
    "current",
    "status",
    "set_collision_margin",
]


def check_positioners(positioner_ids, command, fps, initialised=False):
    """Checks if some of the positioners are not connected."""

    if any([pid not in fps.positioners for pid in positioner_ids]):
        command.fail(error="some positioners are not connected.")
        return False

    if initialised:
        if any([not fps[pid].initialised for pid in positioner_ids]):
            command.fail(error="some positioners are not initialised.")
            return False

    return True


@jaeger_parser.command()
@click.argument("POSITIONER-IDS", type=int, nargs=-1)
@click.argument("ALPHA", type=click.FloatRange(-10.0, 370.0), required=False)
@click.argument("BETA", type=click.FloatRange(-10.0, 370.0), required=False)
@click.option(
    "-l",
    "--from-file",
    type=click.Path(exists=True, dir_okay=False),
    help="Loads the trajectory from a JSON trajectory file. Unless --go-cowboy is used "
    "only the end points of the trajectories for each positioner are used.",
)
@click.option(
    "-r",
    "--relative",
    is_flag=True,
    help="Whether this is a relative move",
)
@click.option(
    "-s",
    "--speed",
    type=click.FloatRange(100.0, 4000.0),
    help="The speed for both alpha and beta arms, in RPS on the input.",
)
@click.option(
    "-a",
    "--all",
    is_flag=True,
    default=False,
    help="Applies to all valid positioners.",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Forces a move to happen.",
)
@click.option(
    "--use-sync/-no-use-sync",
    " /-S",
    default=None,
    help="Whether to use the SYNC line to start the trajectory.",
)
@click.option(
    "--go-cowboy",
    is_flag=True,
    help="If set, does not use kaiju-validated trajectories.",
)
async def goto(
    command: Command[JaegerActor],
    fps: FPS,
    positioner_ids: tuple[int, ...] | list[int],
    alpha: float | None,
    beta: float | None,
    from_file: str | None,
    speed: float | None,
    all: bool = False,
    force: bool = False,
    relative: bool = False,
    use_sync: bool = True,
    go_cowboy: bool = False,
):
    """Sends positioners to a given (alpha, beta) position."""

    assert command.actor

    if from_file is None and (alpha is None or beta is None):
        return command.fail(error="alpha and beta or --from-file are required.")

    if fps.locked:
        return command.fail(error="FPS is locked. Cannot send goto.")

    if fps.moving:
        return command.fail(error="FPS is moving. Cannot send goto.")

    if from_file is not None:
        trajectory = json.loads(open(from_file, "r").read())["trajectory"]

        if go_cowboy is True:
            try:
                await send_trajectory(
                    fps,
                    trajectory,
                    use_sync_line=use_sync,
                    command=command,
                    extra_dump_data={"kaiju_trajectory": False},
                )
                return command.finish()
            except (JaegerError, TrajectoryError) as err:
                return command.fail(error=f"Goto command failed: {err}")

        else:
            new_positions = {}
            for positioner_id in trajectory:
                alpha = trajectory[positioner_id]["alpha"][-1][0]
                beta = trajectory[positioner_id]["beta"][-1][0]
                new_positions[int(positioner_id)] = (alpha, beta)

    else:

        assert alpha is not None and beta is not None

        if all:
            if not force:
                return command.fail("Use --force to move all positioners at once.")
            positioner_ids = list(fps.positioners.keys())
        else:
            positioner_ids = list(positioner_ids)

        if not relative:
            if alpha < 0 or beta < 0:
                return command.fail("Negative angles only allowed in relative mode.")

        if not check_positioners(positioner_ids, command, fps, initialised=True):
            return

        new_positions = {}
        for pid in positioner_ids:
            new_positions[pid] = (alpha, beta)

    try:
        await goto_(
            fps,
            new_positions,
            speed=speed,
            relative=relative,
            use_sync_line=use_sync,
            command=command,
            go_cowboy=go_cowboy,
        )
    except (JaegerError, TrajectoryError) as err:
        return command.fail(error=f"Goto command failed: {err}")

    command.finish()


@jaeger_parser.command()
@click.argument("POSITIONER-ID", type=int, nargs=-1)
@click.argument("ALPHA", type=click.FloatRange(50.0, 5000.0))
@click.argument("BETA", type=click.FloatRange(50.0, 5000.0))
@click.option(
    "-a",
    "--all",
    is_flag=True,
    default=False,
    help="applies to all valid positioners.",
)
async def speed(command, fps, positioner_id, alpha, beta, all):
    """Sets the ``(alpha, beta)`` speed in RPM on the input."""

    if all:
        positioner_id = list(fps.positioners.keys())

    if not check_positioners(positioner_id, command, fps, initialised=True):
        return

    if fps.moving:
        return command.fail(error="FPS is moving. Cannot send set_speed.")

    tasks = []
    for pid in positioner_id:
        tasks.append(fps.positioners[pid].set_speed(alpha, beta))

    result = await clu.as_complete_failer(tasks, on_fail_callback=fps.stop_trajectory)

    if not result[0]:
        error_message = result[1] or "set speed command failed"
        command.set_status(clu.CommandStatus.FAILED, error=error_message)
    else:
        command.set_status(clu.CommandStatus.DONE, text="Set speed done")


@jaeger_parser.command()
@click.argument("POSITIONER-ID", type=int, nargs=-1)
@click.option(
    "--datums",
    is_flag=True,
    help="If set, initialises the datums.",
)
async def initialise(command, fps, positioner_id, datums=False):
    """Initialises positioners."""

    if not check_positioners(positioner_id, command, fps):
        return

    tasks = []
    for pid in positioner_id:
        tasks.append(fps.positioners[pid].initialise(initialise_datums=datums))

    result = await clu.as_complete_failer(tasks)

    if not result[0]:
        error_message = result[1] or "initialise failed"
        command.set_status(clu.CommandStatus.FAILED, error=error_message)
    else:
        command.set_status(clu.CommandStatus.DONE, text="Initialisation complete")


@jaeger_parser.command()
@click.argument("POSITIONERS", type=int, nargs=-1, required=False)
@click.option(
    "-q",
    "--quiet",
    is_flag=True,
    help="Do not print the status of each positioner.",
)
async def status(
    command: JaegerCommandType,
    fps: FPS,
    positioners,
    quiet: bool = False,
):
    """Reports the position and status bit of a list of positioners."""

    positioner_ids = positioners or list(fps.positioners.keys())
    actor = command.actor

    if not check_positioners(positioner_ids, command, fps):
        return

    if len(positioners) == 0:
        command.actor.write("d", {"alive_at": time()}, broadcast=True)
        command.info(locked=fps.locked)
        command.info(folded=(await fps.is_folded()))
        command.info(n_positioners=len(fps.positioners))
        command.info(fps_status=f"0x{fps.status.value:x}")
        command.info(message={k: int(v) for k, v in actor.alerts.keywords.items()})

    try:
        await fps.update_status(positioner_ids=0)
        await fps.update_position(positioner_ids=positioner_ids)
    except JaegerError as err:
        return command.fail(error=f"Failed reporting status: {err}")

    if quiet is False:
        for pid in sorted(positioner_ids):
            p = fps[pid]

            alpha_pos = -999 if p.alpha is None else numpy.round(p.alpha, 4)
            beta_pos = -999 if p.beta is None else numpy.round(p.beta, 4)

            n_trajs_pid = "?"

            if pid in fps.positioner_to_bus and isinstance(fps.can, JaegerCAN):
                interface, bus = fps.positioner_to_bus[pid]
                interface = fps.can.interfaces.index(interface) + 1
            else:
                interface = -1
                bus = -1

            command.write(
                "i",
                positioner_status=[
                    p.positioner_id,
                    alpha_pos,
                    beta_pos,
                    f"0x{int(p.status):x}",
                    p.initialised,
                    p.disabled,
                    p.offline,
                    p.is_bootloader() or False,
                    p.firmware or "?",
                    interface,
                    bus,
                    n_trajs_pid,
                ],
            )

    await clu.Command("ieb status", parent=command).parse()

    command.set_status(clu.CommandStatus.DONE)


@jaeger_parser.command()
@click.argument("POSITIONER-ID", type=int, nargs=-1, required=False)
@click.argument("ALPHA", type=click.FloatRange(0.0, 100.0))
@click.argument("BETA", type=click.FloatRange(0.0, 100.0))
@click.option(
    "-a",
    "--all",
    is_flag=True,
    default=False,
    help="applies to all connected positioners.",
)
async def current(command, fps, positioner_id, alpha, beta, all):
    """Sets the current of the positioner."""

    if all:
        positioner_id = [pid for pid in fps.positioners if fps[pid].initialised]

    if len(positioner_id) == 0:
        return command.fail(error="no positioners provided.")

    if not check_positioners(positioner_id, command, fps):
        return

    if fps.moving:
        return command.fail(error="FPS is moving. Cannot send set current.")

    commands = [
        fps.send_command(SetCurrent(pid, alpha=alpha, beta=beta))
        for pid in positioner_id
    ]
    await asyncio.gather(*commands)

    return command.finish(text="current changed.")


@jaeger_parser.command()
async def stop(command, fps):
    """Stops the positioners and clear flags."""

    await fps.stop_trajectory()
    await fps.update_status(timeout=0.1)
    await fps.update_position()

    command.set_status(clu.CommandStatus.DONE, text="Trajectory aborted")


@jaeger_parser.command()
async def unlock(command, fps):
    """Unlocks the FPS."""

    result = await fps.unlock()

    if result:
        command.info(locked=False)
        return command.finish(text="FPS unlocked")
    else:
        return command.fail(error="failed to unlock FPS")


@jaeger_parser.command()
@click.argument("PATH", type=str)
async def trajectory(command, fps, path):
    """Sends a trajectory from a file."""

    if fps.moving:
        return command.fail(error="FPS is moving. Cannot send trajectory.")

    if fps.locked:
        return command.fail(
            error=f"FPS is locked by {fps.locked_by}. Cannot send trajectory."
        )

    path = pathlib.Path(path).expanduser()
    if not path.exists():
        raise click.BadParameter(f"path {path!s} does not exist.")

    try:

        trajectory = Trajectory(fps, path)

        command.debug(text="sending trajectory ...")
        await trajectory.send()
        if trajectory.failed:
            return command.fail(error="failed sending trajectory with unknown error.")

        command.debug(text=f"trajectory sent in {trajectory.data_send_time:.2f} s.")
        command.info(
            text=f"move will take {trajectory.move_time:.2f} s",
            move_time=f"{trajectory.move_time:.2f}",
        )

        await trajectory.start()
        if trajectory.failed:
            return command.fail(error="failed starting trajectory with unknown error.")

        return command.finish(text="trajectory completed.")

    except TrajectoryError as ee:
        return command.fail(error=str(ee))


@jaeger_parser.group()
def hall():
    """Turns the hall sensor on/off."""

    pass


@hall.command()
@click.argument("POSITIONER-ID", type=int, nargs=-1, required=False)
async def on(command, fps, positioner_id):
    """Turns the hall sensor on."""

    if positioner_id is None:
        positioner_id = list(fps.positioners.keys())

    if not check_positioners(positioner_id, command, fps, initialised=False):
        return

    command.debug("Turning hall sensors ON")
    await fps.send_command("HALL_ON", positioner_ids=positioner_id)

    command.debug("Waiting 5 seconds ...")
    await asyncio.sleep(5)

    command.finish()


@hall.command()
@click.argument("POSITIONER-ID", type=int, nargs=-1, required=False)
async def off(command, fps, positioner_id):
    """Turns the hall sensor off."""

    if positioner_id is None:
        positioner_id = list(fps.positioners.keys())

    if not check_positioners(positioner_id, command, fps, initialised=False):
        return

    command.debug("Turning hall sensors OFF")
    await fps.send_command("HALL_OFF", positioner_ids=positioner_id)

    command.finish()


@jaeger_parser.group()
def led():
    """Turns the positioner LED on/off."""

    pass


@led.command(name="on")
@click.argument("POSITIONER-ID", type=int, nargs=-1, required=False)
async def led_on(command, fps, positioner_id):
    """Turns the LED on."""

    if positioner_id is None:
        positioner_id = list(fps.positioners.keys())

    if not check_positioners(positioner_id, command, fps, initialised=False):
        return

    command.debug("Turning LED on")
    await fps.send_command("SWITCH_LED_ON", positioner_ids=positioner_id)

    command.finish()


@led.command(name="off")
@click.argument("POSITIONER-ID", type=int, nargs=-1, required=False)
async def led_off(command, fps, positioner_id):
    """Turns the LED off."""

    if positioner_id is None:
        positioner_id = list(fps.positioners.keys())

    if not check_positioners(positioner_id, command, fps, initialised=False):
        return

    command.debug("Turning LED off")
    await fps.send_command("SWITCH_LED_OFF", positioner_ids=positioner_id)

    command.finish()


@jaeger_parser.command()
async def reload(command, fps):
    """Reinitialise the FPS."""

    try:
        await fps.initialise(start_pollers=fps.pollers.running)
    except BaseException as err:
        return command.fail(error=f"Initialisation failed: {err}")

    return command.finish(text="FPS was reinitialised.")


@jaeger_parser.command(name="set-collision-margin")
@click.argument("MARGIN", type=click.IntRange(-30, 30))
@click.option(
    "-p",
    "--positioners",
    type=str,
    help="Comma-separated positioners to which to apply the margin. "
    "If not set, applies to all the positioners.",
)
async def set_collision_margin(
    command: JaegerCommandType,
    fps: FPS,
    margin: int,
    positioners: str | None = None,
):
    """Change the collision margin. The collision margin must be -30 to 30 degrees."""

    if positioners is not None:
        positioner_ids = list(map(int, positioners.split(",")))
    else:
        positioner_ids = None

    margin_command = await fps.send_command(
        "SET_INCREASE_COLLISION_MARGIN",
        positioner_ids=positioner_ids,
        margin=margin,
    )

    if margin_command.status.failed:
        return command.fail("Failed updating collision margin.")

    return command.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2018-10-03
# @Filename: goto.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import warnings

from typing import TYPE_CHECKING, Dict, List, Optional, Tuple

import jaeger
from jaeger import config
from jaeger.commands import Command, CommandID
from jaeger.exceptions import (
    FPSLockedError,
    JaegerError,
    JaegerUserWarning,
    TrajectoryError,
)
from jaeger.kaiju import get_path_pair
from jaeger.utils import (
    bytes_to_int,
    get_goto_move_time,
    int_to_bytes,
    motor_steps_to_angle,
)
from jaeger.utils.helpers import run_in_executor

from .trajectory import send_trajectory


if TYPE_CHECKING:
    from clu.command import Command as CluCommand

    from jaeger.actor import JaegerActor
    from jaeger.fps import FPS


__all__ = [
    "GoToDatums",
    "GoToDatumAlpha",
    "GoToDatumBeta",
    "GotoAbsolutePosition",
    "GotoRelativePosition",
    "SetActualPosition",
    "SetSpeed",
    "SetCurrent",
    "goto",
]


TIME_STEP = jaeger.config["positioner"]["time_step"]


class GoToDatums(Command):
    """Initialises and zeroes the positioner."""

    command_id = CommandID.GO_TO_DATUMS
    broadcastable = False
    move_command = True


class GoToDatumAlpha(Command):
    """Initialises and zeroes the alpha arm of the positioner."""

    command_id = CommandID.GO_TO_DATUM_ALPHA
    broadcastable = False
    move_command = True


class GoToDatumBeta(Command):
    """Initialises and zeroes the beta arm of the positioner."""

    command_id = CommandID.GO_TO_DATUM_BETA
    broadcastable = False
    move_command = True


class GotoAbsolutePosition(Command):
    """Moves alpha and beta to absolute positions in degrees."""

    command_id = CommandID.GO_TO_ABSOLUTE_POSITION
    broadcastable = False
    move_command = True

    def __init__(
        self,
        positioner_ids: int | List[int],
        alpha: float | None = None,
        beta: float | None = None,
        **kwargs,
    ):

        if alpha is not None and beta is not None:
            alpha_steps, beta_steps = motor_steps_to_angle(alpha, beta, inverse=True)

            alpha_bytes = int_to_bytes(alpha_steps, dtype="i4")
            beta_bytes = int_to_bytes(beta_steps, dtype="i4")

            data = alpha_bytes + beta_bytes
            kwargs["data"] = data

        super().__init__(positioner_ids, **kwargs)

    @staticmethod
    def decode(data):
        """Decodes message data into alpha and beta moves."""

        alpha_steps = bytes_to_int(data[0:4], dtype="i4")
        beta_steps = bytes_to_int(data[4:8], dtype="i4")

        return motor_steps_to_angle(alpha_steps, beta_steps)

    def get_replies(self) -> Dict[int, Tuple[float, float]]:
        return self.get_move_time()

    def get_move_time(self):
        """Returns the time needed to move to the commanded position.

        Raises
        ------
        ValueError
            If no reply has been received or the data cannot be parsed.

        """

        move_times = {}
        for reply in self.replies:
            data = reply.data

            alpha = bytes_to_int(data[0:4], dtype="i4")
            beta = bytes_to_int(data[4:], dtype="i4")

            move_times[reply.positioner_id] = [alpha * TIME_STEP, beta * TIME_STEP]

        return move_times


class GotoRelativePosition(GotoAbsolutePosition):
    """Moves alpha and beta a relative number of degrees."""

    command_id = CommandID.GO_TO_RELATIVE_POSITION
    broadcastable = False
    move_command = True


class SetActualPosition(Command):
    """Sets the current position of the alpha and beta arms."""

    command_id = CommandID.SET_ACTUAL_POSITION
    broadcastable = False
    safe = True
    move_command = True  # Technically not a move command but we don't
    # want to issue it during a move.

    def __init__(
        self,
        positioner_ids: int | List[int],
        alpha: float | None = None,
        beta: float | None = None,
        **kwargs,
    ):

        if alpha is not None and beta is not None:

            alpha_steps, beta_steps = motor_steps_to_angle(alpha, beta, inverse=True)

            alpha_bytes = int_to_bytes(int(alpha_steps), dtype="i4")
            beta_bytes = int_to_bytes(int(beta_steps), dtype="i4")

            data = alpha_bytes + beta_bytes

            kwargs["data"] = data

        super().__init__(positioner_ids, **kwargs)


class SetSpeed(Command):
    """Sets the speeds of the alpha and beta motors."""

    command_id = CommandID.SET_SPEED
    broadcastable = False
    safe = True
    move_command = False

    def __init__(
        self,
        positioner_ids: int | List[int],
        alpha: float | None = None,
        beta: float | None = None,
        **kwargs,
    ):

        if alpha is not None and beta is not None:
            assert alpha >= 0 and beta >= 0, "invalid speed."

            data = int_to_bytes(int(alpha)) + int_to_bytes(int(beta))
            kwargs["data"] = data

        super().__init__(positioner_ids, **kwargs)

    @staticmethod
    def encode(alpha, beta):
        """Encodes the alpha and beta speed as bytes."""

        data_speed = int_to_bytes(int(alpha)) + int_to_bytes(int(beta))

        return data_speed


class SetCurrent(Command):
    """Sets the current of the alpha and beta motors."""

    command_id = CommandID.SET_CURRENT
    broadcastable = False
    safe = True
    move_command = True

    def __init__(
        self,
        positioner_ids: int | List[int],
        alpha: float | None = None,
        beta: float | None = None,
        **kwargs,
    ):

        if alpha is not None and beta is not None:
            assert alpha >= 0 and beta >= 0, "invalid current."

            data = int_to_bytes(int(alpha)) + int_to_bytes(int(beta))
            kwargs["data"] = data

        super().__init__(positioner_ids, **kwargs)


async def goto(
    fps: FPS,
    new_positions: dict[int, tuple[float, float]],
    speed: Optional[float] = None,
    relative: bool = False,
    use_sync_line: bool | None = None,
    go_cowboy: bool = False,
    force: bool = False,
    command: CluCommand[JaegerActor] | None = None,
):
    """Send positioners to a given position using a trajectory with ``kaiju`` check.

    Parameters
    ----------
    fps
        The `.FPS` instance.
    new_positions
        The new positions as a dictionary of positioner ID to a tuple of new
        alpha and beta angles. Positioners not specified will be kept on the
        same positions.
    speed
        The speed to use.
    relative
        If `True`, ``alpha`` and ``beta`` are considered relative angles.
    use_sync_line
        Whether to use the SYNC line to start the trajectories.
    go_cowboy
        If set, does not create a ``kaiju``-safe trajectory. Use at your own risk.
    force
        If ``go_cowboy=False`` and the trajectory is deadlocked, a `.TrajectoryError`
        will be raised. Use ``force=True`` to apply the trajectory anyway.
    command
        A command to pass to `.send_trajectory` to output additional information.

    """

    if fps.locked:
        FPSLockedError("The FPS is locked.")

    if fps.moving:
        raise JaegerError("FPS is moving. Cannot send goto.")

    for pid in new_positions:
        if pid not in fps.positioners:
            raise JaegerError(f"Positioner ID {pid} is not connected.")

    speed = float(speed or config["positioner"]["motor_speed"])
    if speed < 500 or speed > 5000:
        raise JaegerError("Invalid speed.")

    positioner_ids = list(new_positions.keys())
    await fps.update_position(positioner_ids=positioner_ids)

    trajectories = {}

    if go_cowboy is True:
        for pid in positioner_ids:
            pos = fps[pid]

            if pos.alpha is None or pos.beta is None:
                raise JaegerError(
                    f"Positioner {pid}: cannot goto with unknown position."
                )

            current_alpha = pos.alpha
            current_beta = pos.beta

            if relative is True:
                alpha_end = current_alpha + new_positions[pid][0]
                beta_end = current_beta + new_positions[pid][1]
            else:
                alpha_end = new_positions[pid][0]
                beta_end = new_positions[pid][1]

            alpha_delta = abs(alpha_end - current_alpha)
            beta_delta = abs(beta_end - current_beta)

            time_end = [
                get_goto_move_time(alpha_delta, speed=speed),
                get_goto_move_time(beta_delta, speed=speed),
            ]

            trajectories[pid] = {
                "alpha": [(current_alpha, 0.1), (alpha_end, time_end[0] + 0.1)],
                "beta": [(current_beta, 0.1), (beta_end, time_end[1] + 0.1)],
            }

    else:
        if relative is True:
            raise JaegerError("relative is not implemented for kaiju moves.")

        data = {"collision_buffer": None, "grid": {}}

        for pid, (current_alpha, current_beta) in fps.get_positions_dict().items():

            if current_alpha is None or current_beta is None:
                raise JaegerError(f"Positioner {pid} does not know its position.")

            if pid in new_positions:
                data["grid"][int(pid)] = (
                    current_alpha,
                    current_beta,
                    new_positions[pid][0],
                    new_positions[pid][1],
                    fps[pid].disabled,
                )
            else:
                data["grid"][int(pid)] = (
                    current_alpha,
                    current_beta,
                    current_alpha,
                    current_beta,
                    fps[pid].disabled,
                )

        (to_destination, _, did_fail, deadlocks) = await run_in_executor(
            get_path_pair,
            data=data,
            path_generation_mode="greedy",
            stop_if_deadlock=force,
            executor="process",
            ignore_did_fail=force,
        )

        if did_fail is True:
            if force is False:
                raise TrajectoryError(
                    "Cannot execute trajectory. Found "
                    f"{len(deadlocks)} deadlocks ({deadlocks})."
                )
            else:
                warnings.warn(
                    f"Found {len(deadlocks)} deadlocks but applying trajectory.",
                    JaegerUserWarning,
                )

        trajectories = to_destination

    return await send_trajectory(
        fps,
        trajectories,
        use_sync_line=use_sync_line,
        command=command,
        extra_dump_data={"kaiju_trajectory": not go_cowboy},
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2018-08-27
# @Filename: can.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import binascii
import collections
import pprint
import re
import socket
from dataclasses import dataclass, field

from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generic,
    List,
    Optional,
    Type,
    TypeVar,
    cast,
)

import jaeger
from jaeger import can_log, config, log, start_file_loggers
from jaeger.commands import Command, CommandID, EmptyPool
from jaeger.exceptions import JaegerCANError
from jaeger.interfaces import BusABC, CANNetBus, Message, Notifier, VirtualBus
from jaeger.maskbits import CommandStatus
from jaeger.utils import Poller, parse_identifier


try:
    from can.interfaces.slcan import slcanBus  # type: ignore
    from can.interfaces.socketcan import SocketcanBus  # type: ignore
except ImportError:
    SocketcanBus = None
    slcanBus = None

if TYPE_CHECKING:
    from .fps import FPS


__all__ = ["JaegerCAN", "CANnetInterface", "INTERFACES"]


LOG_HEADER = "({cmd.command_id.name}, {cmd.command_uid}):"

#: Accepted CAN interfaces and whether they are multibus.
INTERFACES = {
    "slcan": {"class": slcanBus, "multibus": False},
    "socketcan": {"class": SocketcanBus, "multibus": False},
    "virtual": {"class": VirtualBus, "multibus": False},
    "cannet": {"class": CANNetBus, "multibus": True},
}


Bus_co = TypeVar("Bus_co", bound="BusABC")
T = TypeVar("T", bound="JaegerCAN")


@dataclass
class JaegerCAN(Generic[Bus_co]):
    """A CAN interface with a command queue and reply handling.

    Provides support for multi-channel CAN networks, with each channel being
    able to host more than one bus. The recommended way to instantiate a new
    `.JaegerCAN` object is using the `.create` classmethod ::

        can = await JaegerCAN.create(...)

    which is equivalent to ::

        can = JaegerCAN(...)
        await can.start()

    Parameters
    ----------
    interface_type
        One of `~jaeger.can.INTERFACES`.
    channels
        A list of channels to be used to instantiate the interfaces.
    fps
        The focal plane system.
    interface_args
        Keyword arguments to pass to the interfaces when
        initialising it (e.g., port, baudrate, etc).

    """

    interface_type: str
    channels: list | tuple
    fps: Optional[jaeger.fps.FPS] = None
    interface_args: Dict[str, Any] = field(default_factory=dict)

    def __post_init__(self):

        if self.interface_type not in INTERFACES:
            raise ValueError(f"Invalid interface {self.interface_type}.")

        # Start can file logger
        if config["debug"] is True:
            start_file_loggers(start_log=False, start_can=True)

        # List of interfaces for each channel.
        self.interfaces: List[Bus_co] = []

        self.multibus: bool = False

        self._started: bool = False

        # Currently running commands.
        self.running_commands: Dict[int, Command] = {}

        self.command_queue: asyncio.Queue[Command] | None = None
        self._command_queue_task: asyncio.Task | None = None

        self.notifier: Notifier | None = None

    async def start(self: T) -> T:

        self.stop()

        itype = self.interface_type

        InterfaceClass: Type[Bus_co] = INTERFACES[itype]["class"]
        self.multibus = INTERFACES[itype]["multibus"]

        if not isinstance(self.channels, (list, tuple)):
            self.channels = [self.channels]

        for channel in self.channels:
            iargs = "".join([f" {k}={repr(v)}" for k, v in self.interface_args.items()])
            log.debug(f"creating interface {itype}, channel={channel!r}{iargs}.")
            try:
                interface = InterfaceClass(channel, **self.interface_args)
                result = await interface.open()
                if result is False:
                    raise ConnectionError()
                self.interfaces.append(interface)
            except ConnectionResetError:
                log.error(
                    f"connection to {itype}:{channel} failed. "
                    "Possibly another instance is connected to the device."
                )
            except (socket.timeout, ConnectionError, ConnectionRefusedError, OSError):
                log.error(f"connection to {itype}:{channel} failed.")
            except Exception as ee:
                raise ee.__class__(f"connection to {itype}:{channel} failed: {ee}.")

        self.command_queue = asyncio.Queue()
        self._command_queue_task = asyncio.create_task(self._process_command_queue())

        self.notifier = Notifier(
            listeners=[self._process_reply_queue],
            buses=self.interfaces,
        )

        self._started = True

        return self

    def stop(self):
        """Stops the interfaces."""

        if self.notifier:
            self.notifier.stop()
            self.notifier = None

        for interface in self.interfaces:
            interface: Any
            try:
                interface.close()
            except AttributeError:
                pass

        self.interfaces = []

        if self._command_queue_task:
            self._command_queue_task.cancel()

        self._started = False

    @classmethod
    async def create(
        cls,
        profile: Optional[str] = None,
        fps: FPS | None = None,
        interface_type: Optional[str] = None,
        channels: list | tuple = [],
        interface_args: Dict[str, Any] = {},
    ) -> "JaegerCAN":
        """Create and initialise a new bus interface from a configuration profile.

        This is the preferred method to initialise a new `.JaegerCAN` instance and is
        equivalent to calling ``JaegerCAN`` and then `~.JaegerCAN.start`.

        Parameters
        ----------
        profile
            The name of the profile that defines the bus interface, or `None`
            to use the default configuration.
        fps
            The focal plane system.
        interface_type
            One of `~jaeger.can.INTERFACES`. Cannot be used with ``profile``.
        channels
            A list of channels to be used to instantiate the interfaces.
        interface_args
            Keyword arguments to pass to the interfaces when
            initialising it (e.g., port, baudrate, etc).

        """

        if profile is not None or (profile is None and interface_type is None):
            if "profiles" not in config:
                raise ValueError("No 'interfaces' section in the configuration file.")

            if profile is None:
                if "default" not in config["profiles"]:
                    raise ValueError("Default interface not defined in configuration.")
                profile = config["profiles"]["default"]

            if profile not in config["profiles"]:
                raise ValueError(f"Invalid interface profile {profile}")

            config_data = config["profiles"][profile].copy()

            interface_type = config_data.pop("interface")
            if interface_type not in INTERFACES:
                raise ValueError(f"invalid interface {interface_type}")

            if "channel" in config_data:
                channels = [config_data.pop("channel")]
            elif "channels" in config_data:
                channels = config_data.pop("channels")
                assert isinstance(channels, (list, tuple)), "channels must be a list"
            else:
                raise KeyError("channel or channels key not found.")

            interface_args = config_data

        elif profile is not None and interface_type is not None:
            raise JaegerCANError("profile and interface_type are mutually exclusive.")

        if interface_type == "cannet":
            cls = CANnetInterface

        assert interface_type, "interface_type not set. This should not have happened."

        instance = cls(
            interface_type,
            channels=channels,
            fps=fps,
            interface_args=interface_args,
        )

        await instance.start()

        return instance

    def refresh_running_commands(self):
        """Clears completed commands."""

        rc = self.running_commands
        self.running_commands = {key: cmd for key, cmd in rc.items() if not cmd.done()}

    async def _process_command_queue(self):
        """Processes messages in the command queue."""

        assert self.command_queue

        while True:

            cmd = await self.command_queue.get()

            log_header = LOG_HEADER.format(cmd=cmd)

            if cmd.status != CommandStatus.READY:
                if cmd.status != CommandStatus.CANCELLED:
                    can_log.error(
                        f"{log_header} command is not ready "
                        f"(status={cmd.status.name!r})"
                    )
                    cmd.cancel()
                continue

            try:
                self.send_messages(cmd)
            except EmptyPool:
                if cmd.positioner_ids == [0]:
                    # We'll ignore this case silently since generally this happens only
                    # with GET_FIRMWARE or GET_STATUS that can usually be delayed.
                    loop = asyncio.get_event_loop()
                    loop.call_later(1, self.command_queue.put_nowait, cmd)
                    continue
            except jaeger.JaegerError as ee:
                can_log.error(f"found error while getting messages: {ee}")
                continue

    async def _process_reply_queue(self, msg: Message):
        """Processes one reply message."""

        positioner_id, command_id, reply_uid, __ = parse_identifier(msg.arbitration_id)

        if command_id == CommandID.COLLISION_DETECTED:

            # Sending stop trajectories causes many more robots to report a collision
            # so if the FPS has already been locked we ignore those.
            if not self.fps or self.fps.locked:
                return

            log.error(
                f"A collision was detected in positioner {positioner_id}. "
                "Sending SEND_TRAJECTORY_ABORT and locking the FPS."
            )

            if self.fps:
                await self.fps.lock(by=[positioner_id])
                return

        if command_id == 0:
            can_log.warning(
                "invalid command with command_id=0, "
                f"arbitration_id={msg.arbitration_id} received. "
                "Ignoring it."
            )
            return

        command_id_flag = CommandID(command_id)

        self.refresh_running_commands()

        cmd_key = (positioner_id << 25) + (command_id << 15) + reply_uid

        if cmd_key in self.running_commands:
            running_cmd = self.running_commands[cmd_key]
        elif (command_id << 15) + reply_uid in self.running_commands:
            # Checks if the reply corresponds to a broadcast.
            cmd_key = (command_id << 15) + reply_uid
            running_cmd = self.running_commands[cmd_key]
        else:
            can_log.debug(
                f"[{command_id_flag.name}, {positioner_id}]: "
                f"cannot find a matching running command."
            )
            return

        if not running_cmd.is_broadcast:
            if reply_uid not in running_cmd.message_uids:
                can_log.debug(
                    f"[{command_id_flag.name}, {positioner_id}]: "
                    f"matching command does not contain reply UID={reply_uid}."
                )
                return

        can_log.debug(
            f"[{command_id_flag.name}, "
            f"{positioner_id}, {running_cmd.command_uid}]: "
            f"queuing reply UID={reply_uid} "
            f"to command {running_cmd.command_uid}."
        )

        asyncio.create_task(running_cmd.process_reply(msg))

    def send_messages(self, cmd: Command):
        """Sends messages to the interface.

        This method exists separate from _process_queue so that it can be used
        to send command messages to the interface synchronously.

        """

        log_header = LOG_HEADER.format(cmd=cmd)

        if cmd.status != CommandStatus.READY:
            if cmd.status != CommandStatus.CANCELLED:
                can_log.error(
                    f"{log_header} command is not ready "
                    f"(status={cmd.status.name!r})"
                )
                cmd.cancel()
            return

        can_log.debug(
            f"{log_header} sending command {cmd.command_uid} "
            f"to positioners {cmd.positioner_ids!r}."
        )
        can_log.debug(log_header + " sending messages to CAN bus.")

        messages = cmd.get_messages()

        for message in messages:

            if cmd.status.failed:
                can_log.debug(
                    f"{log_header} not sending more messages "
                    "since this command has failed."
                )
                break

            cmd_key = message.positioner_id << 25
            cmd_key += message.command.command_id << 15
            cmd_key += message.uid

            self.running_commands[cmd_key] = message.command

            # Get the interface and buses to which to send this command.
            interfaces = self.interfaces
            bus = None
            is_multibus = self.multibus or len(self.interfaces) > 1
            if is_multibus and message.positioner_id != 0:
                if self.fps and message.positioner_id in self.fps.positioner_to_bus:
                    interface, bus = self.fps.positioner_to_bus[message.positioner_id]
                    interfaces = [cast(Bus_co, interface)]

            for iface in interfaces:

                iface_idx = self.interfaces.index(iface)
                data_hex = binascii.hexlify(message.data).decode()
                can_log.debug(
                    log_header + "sending message with "
                    f"arbitration_id={message.arbitration_id}, "
                    f"UID={message.uid}, "
                    f"and data={data_hex!r} to "
                    f"interface {iface_idx}, "
                    f"bus={0 if not bus else bus!r}."
                )

                if bus:
                    iface.send(message, bus=bus)  # type: ignore
                else:
                    iface.send(message)

        cmd.status = CommandStatus.RUNNING

    @staticmethod
    def print_profiles() -> List[str]:
        """Prints interface profiles and returns a list of profile names."""

        pprint.pprint(config["interfaces"])

        return list(config["interfaces"].keys())


CANNET_ERRORS = {
    0: "Unknown error <error_code>",
    1: "CAN <port_num> baud rate not found",
    2: "CAN <port_num> stop failed",
    3: "CAN <port_num> start failed",
    4: "CAN <port_num> extended filter is full",
    5: "CAN <port_num> standard open filter set twice",
    6: "CAN <port_num> standard filter is full",
    7: "CAN <port_num> invalid identifier or mask for filter add",
    8: "CAN <port_num> baud rate detection is busy",
    9: "CAN <port_num> invalid parameter type",
    10: "CAN <port_num> invalid CAN state",
    11: "CAN <port_num> invalid parameter mode",
    12: "CAN <port_num> invalid port number",
    13: "CAN <port_num> init auto baud failed",
    14: "CAN <port_num> filter parameter is missing",
    15: "CAN <port_num> bus off parameter is missing",
    16: "CAN <port_num> parameter is missing",
    17: "DEV parameter is missing",
    18: "CAN <port_num> invalid parameter brp",
    19: "CAN <port_num> invalid parameter sjw",
    20: "CAN <port_num> invalid parameter tSeg1",
    21: "CAN <port_num> invalid parameter tSeg2",
    22: "CAN <port_num> init custom failed",
    23: "CAN <port_num> init failed",
    24: "CAN <port_num> reset failed",
    25: "CAN <port_num> filter parameter is missing",
    27: "CYC parameter is missing",
    28: "CYC message <msg_num> stop failed",
    29: "CYC message <msg_num> init failed",
    30: "CYC message <msg_num> invalid parameter port",
    31: "CYC message <msg_num> invalid parameter msg_num",
    32: "CYC message <msg_num> invalid parameter time",
    33: "CYC message <msg_num> invalid parameter data",
}


@dataclass
class CANnetInterface(JaegerCAN[CANNetBus]):
    r"""An interface class specifically for the CAN\@net 200/420 device.

    This class bahaves as `.JaegerCAN` but allows communication with the
    device itself and tracks its status.

    """

    status_interval: float = 5

    def __post_init__(self):
        super().__post_init__()
        self._device_status = collections.defaultdict(dict)

        self.device_status_poller: Poller | None = None

    async def start(self):
        r"""Starts CAN\@net connection."""

        await super().start()

        if len(self.interfaces) == 0:
            return

        # Get ID and version of interfaces. No need to ask for this in each
        # status poll.
        for interface in self.interfaces:
            interface.write("DEV IDENTIFY")
            interface.write("DEV VERSION")

        # We use call_soon later to be sure the event loop is running when we start
        # the poller. This prevents problems when using the library in IPython.
        self.device_status_poller = Poller(
            "cannet_device",
            self._get_device_status,
            delay=self.status_interval,
        )
        self.device_status_poller.start()

    def stop(self):
        """Stops the interfaces."""

        super().stop()

        if self.device_status_poller is not None:
            asyncio.create_task(self.device_status_poller.stop())

    async def _process_reply_queue(self, msg: Message):
        """Processes a message checking first if it comes from the device."""

        if msg.arbitration_id == 0:
            return self.handle_device_message(msg)

        await super()._process_reply_queue(msg)

    @property
    def device_status(self):
        """Returns a dictionary with the status of the device."""

        if self.device_status_poller and not self.device_status_poller.running:
            raise ValueError("the device status poller is not running.")

        return self._device_status

    def handle_device_message(self, msg: Message):
        """Handles a reply from the device (i.e., not from the CAN network)."""

        device_status = self._device_status

        interface_id = self.interfaces.index(msg.interface)
        message = msg.data.decode()

        can_log.debug(f"received message {message!r} from interface ID {interface_id}.")

        if message.lower() == "r ok":
            return

        dev_identify = re.match(
            r"^R (?P<device>CAN@net \w+ \d+)$",
            message,
        )
        dev_version = re.match(
            r"^R V(?P<version>(\d+\.*)+)$",
            message,
        )
        dev_error = re.match(
            r"^R ERR (?P<error_code>\d{1,2}) (?P<error_descr>\.+)$",
            message,
        )
        dev_event = re.match(
            r"^E (?P<bus>\d+) (?P<event>.+)$",
            message,
        )
        can_status = re.match(
            r"^R CAN (?P<bus>\d+) (?P<status>[-|\w]{5}) (?P<buffer>\d+)$",
            message,
        )

        if dev_identify:
            device = dev_identify.group("device")
            device_status[interface_id]["name"] = device

        elif dev_version:
            version = dev_version.group("version")
            device_status[interface_id]["version"] = version

        elif dev_error:

            if "errors" not in device_status[interface_id]:
                device_status[interface_id]["errors"] = []

            device_status[interface_id]["errors"].insert(
                0,
                {
                    "error_code": dev_error.group("error_code"),
                    "error_description": dev_error.group("error_descr"),
                    "timestamp": str(msg.timestamp),
                },
            )

        elif dev_event:
            bus, event = dev_event.groups()
            bus = int(bus)

            if "events" not in device_status[interface_id]:
                device_status[interface_id]["events"] = collections.defaultdict(list)

            device_status[interface_id]["events"][bus].insert(
                0, {"event": event, "timestamp": str(msg.timestamp)}
            )

        elif can_status:
            bus, status, buffer = can_status.groups()
            bus = int(bus)
            buffer = int(buffer)

            # Unpack the status characters. If they are different
            # than '-', they are True.
            status_bool = list(map(lambda x: x != "-", status))
            (
                bus_off,
                error_warning,
                data_overrun,
                transmit_pending,
                init_state,
            ) = status_bool

            device_status[interface_id][bus] = {
                "status": status,
                "buffer": buffer,
                "bus_off": bus_off,
                "error_warning_level": error_warning,
                "data_overrun_detected": data_overrun,
                "transmit_pending": transmit_pending,
                "init_state": init_state,
                "timestamp": str(msg.timestamp),
            }

        else:
            can_log.debug(f"message {message!r} cannot be parsed.")

    def _get_device_status(self):
        """Sends commands to the devices to get status updates."""

        for interface in self.interfaces:
            for bus in interface.buses:
                interface.write(f"CAN {bus} STATUS")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-02-02
# @Filename: chiller.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import logging

from typing import TYPE_CHECKING

from jaeger import config
from jaeger.ieb import IEB, Chiller
from jaeger.utils.helpers import BaseBot


if TYPE_CHECKING:
    from jaeger.fps import FPS


__all__ = ["ChillerBot"]


class ChillerBot(BaseBot):
    """Manages the FPS chiller."""

    def __init__(self, fps: FPS):

        self.chiller_config = config.get("chiller", {})
        self.temperature: bool | float = self.chiller_config.get("temperature", False)
        self.flow: bool | float = self.chiller_config.get("flow", False)

        super().__init__(fps)

    async def _loop(self):
        """Sets the chiller set point temperature."""

        chiller = Chiller.create()
        assert chiller is not None

        while True:

            await self._set_temperature(chiller)
            await self._set_flow(chiller)

            await asyncio.sleep(60)

    async def _set_temperature(self, chiller: Chiller):
        """Sets the temperature set point."""

        dev_name = "TEMPERATURE_USER_SETPOINT"
        dev = chiller.get_device(dev_name)

        if not isinstance(self.ieb, IEB) or self.ieb.disabled is True:
            return

        if self.temperature is False or self.temperature is None:
            return

        failed: bool = False

        # Try up to 10 times since sometimes setting the temperature fails.
        for _ in range(10):
            failed = False

            try:
                ambient_temp = (await self.ieb.read_device("T3"))[0]
                rh = (await self.ieb.read_device("RH3"))[0]

                # Dewpoint temperature.
                t_d = ambient_temp - (100 - rh) / 5.0

                current_setpoint = (await dev.read())[0]

                # If we are maintaining a fixed temperature, check if we
                # need to reset the set point and exit.
                if self.temperature is not True and self.temperature != "auto":
                    if abs(current_setpoint - self.temperature) > 0.1:
                        await dev.write(int(self.temperature * 10))
                        self.notify(
                            f"Setting chiller to {self.temperature} C",
                            logging.DEBUG,
                        )
                    break

                # What follows is if we are setting the set point
                # based on the ambient temperature.

                # Calculate the new set point temperature and write it
                # to the device if it's different than the current one.

                # New set point is two degC below ambient clipped to 1 degC
                # and above the dew point depression region.
                new_temp = ambient_temp - 2
                if new_temp <= 1:
                    new_temp = 1
                if new_temp < (t_d + 3):
                    new_temp = t_d + 3

                # Round to closest 0.5
                new_temp = round(new_temp * 2) / 2.0
                delta_temp = abs(current_setpoint - new_temp)
                if delta_temp > 0.1:
                    await dev.write(int(new_temp * 10))
                    self.notify(f"Setting chiller to {new_temp} C", logging.DEBUG)

                break

            except Exception:
                failed = True
                await asyncio.sleep(2)
                continue

        if failed is True:
            self.notify("Failed setting chiller temperature.", level=logging.ERROR)

    async def _set_flow(self, chiller: Chiller):
        """Sets the flow set point."""

        dev_name = "FLOW_USER_SETPOINT"
        dev = chiller.get_device(dev_name)

        if not isinstance(self.ieb, IEB) or self.ieb.disabled is True:
            return

        if self.flow is True or self.flow == "auto":
            self.flow = self.chiller_config.get("flow", False)

        if self.flow is False or self.flow is None:
            return

        failed: bool = False

        # Try up to 10 times since sometimes setting the flow fails.
        for _ in range(10):
            failed = False

            try:
                current_setpoint = (await dev.read())[0]

                if isinstance(self.flow, (float, int)):
                    if abs(current_setpoint - self.flow) > 0.1:
                        await dev.write(int(self.flow * 10))
                        self.notify(
                            f"Setting chiller flow to {self.flow:.1f} gpm.",
                            level=logging.DEBUG,
                        )

                else:
                    self.notify(
                        f"Invalid chiller set point {self.flow!r}.",
                        level=logging.WARNING,
                    )

                break

            except Exception:
                failed = True
                await asyncio.sleep(2)
                continue

        if failed is True:
            self.notify("Failed setting chiller flow.", level=logging.ERROR)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Ricardo Araujo
# @Filename: cannet.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import time

from .bus import BusABC
from .message import Message


class CANNetMessage(Message):
    __slots__ = ("interface", "bus")


class CANNetBus(BusABC):
    r"""Interface for Ixxat CAN\@net NT 200/420.

    Parameters
    ----------
    channel : str
        The IP address of the remote device (e.g. ``192.168.1.1``, ...).
    port : int
        The port of the device.
    bitrate : int
        Bitrate in bit/s.
    buses : list
        The buses to open in the device. Messages that do not specify a
        bus will be sent to all the open buses.
    timeout : float
        Timeout for connection.

    """

    # the supported bitrates and their commands
    _BITRATES = {
        5000: "5",
        10000: "10",
        20000: "20",
        50000: "50",
        62500: "62.5",
        83300: "83.3",
        100000: "100",
        125000: "125",
        500000: "500",
        800000: "800",
        1000000: "1000",
    }

    _REMOTE_PORT = 19228

    LINE_TERMINATOR = b"\n"

    def __init__(
        self,
        channel,
        port=None,
        bitrate=None,
        buses=[1],
        timeout=5,
        **kwargs,
    ):

        if not channel:  # if None or empty
            raise TypeError("Must specify a TCP address.")

        self.channel = channel

        if not bitrate:
            raise TypeError("Must specify a bitrate.")

        self.port = port or self._REMOTE_PORT

        self.bitrate = bitrate
        self.buses = buses

        self.reader: asyncio.StreamReader | None = None
        self.writer: asyncio.StreamWriter | None = None
        self.connected = False

        self._timeout = timeout

        self.channel_info = f"CAN@net channel={channel!r}, buses={self.buses!r}"

        super(CANNetBus, self).__init__(channel, bitrate=None, **kwargs)

    def write(self, string):

        if not self.connected or not self.writer:
            raise ConnectionError(f"Interface {self.channel} is not connected.")

        self.writer.write(string.encode() + self.LINE_TERMINATOR)

    def _write_to_buses(self, string, buses=None):
        """Writes a string to the correct bus."""

        if not buses:
            buses = self.buses
        elif not isinstance(buses, (list, tuple)):
            buses = [buses]

        for bus in buses:
            self.write(string.format(bus=bus))

    async def _open_internal(self, timeout=None):

        timeout = timeout or self._timeout

        self.close()

        try:
            open_conn = asyncio.open_connection(self.channel, self.port)
            self.reader, self.writer = await asyncio.wait_for(open_conn, timeout)
        except asyncio.TimeoutError:
            self.connected = False
            return False

        self.connected = True

        if self.bitrate in self._BITRATES:
            self._write_to_buses("CAN {bus} STOP")
            self._write_to_buses(f"CAN {{bus}} INIT STD {self._BITRATES[self.bitrate]}")
            self._write_to_buses("CAN {bus} FILTER CLEAR")
            self._write_to_buses("CAN {bus} FILTER ADD EXT 00000000 00000000")
        else:
            bitrates = ", ".join(map(str, self._BITRATES))
            raise ValueError(f"Invalid bitrate, choose one of {bitrates}.")

        self._write_to_buses("CAN {bus} START")

        # await self.writer.drain()

        # Clear buffer
        await self.reader.read(8192)

        return True

    def close(self, buses=None):

        if self.writer and not self.writer.is_closing():
            self._write_to_buses("CAN {bus} STOP")
            self.writer.close()

        self.connected = False
        self.writer = self.reader = None

    async def get(self):

        canId = None
        remote = False
        extended = False
        frame = []

        if not self.reader:
            raise ConnectionError(f"Interface {self.channel} is not connected.")

        msgStr = await self.reader.readuntil(self.LINE_TERMINATOR)

        readStr = msgStr.strip(self.LINE_TERMINATOR).decode()
        if not readStr:
            return None

        # Message is M 1 CSD 100 55 AA 55 AA or M 2 CED 18FE0201 01 02 03 04 05 06 07 08
        # Check if we have a message from the CAN network. Otherwise this is a message
        # from the device so we return it.
        data = readStr.split(" ")
        if data[0] != "M":
            msg = CANNetMessage(
                arbitration_id=0,
                timestamp=time.time(),
                dlc=0,
                data=msgStr,
            )
            msg.interface = self
            msg.bus = None
            return msg

        # check if it is the proper CAN bus
        bus = int(data[1])
        if bus not in self.buses:
            return None

        # check if standard packet, FD not supported
        if data[2][0] != "C":
            return None

        # check if remote frame
        if data[2][2] == "D":
            remote = False
        elif data[2][2] == "R":
            remote = True

        # check if standard or extended packet
        if data[2][1] == "S":
            extended = False
        elif data[2][1] == "E":
            extended = True
        else:
            return None

        # get canId
        canId = int(data[3], 16)

        # get frame data
        dlc = 0
        for byte in data[4:]:
            frame.append(int(byte, 16))
            dlc = dlc + 1

        if canId is not None:
            msg = CANNetMessage(
                arbitration_id=canId,
                is_extended_id=extended,
                timestamp=time.time(),
                is_remote_frame=remote,
                dlc=dlc,
                data=frame,
            )
            msg.interface = self
            msg.bus = bus
            return msg

        return None

    def send(self, msg, bus=None):

        buses = bus or self.buses
        if not isinstance(buses, (list, tuple)):
            buses = [buses]

        for bus in buses:

            sendStr = f"M {bus} "

            if msg.is_extended_id:
                if msg.is_remote_frame:
                    sendStr += f"CER {msg.arbitration_id:08X}"
                else:
                    sendStr += f"CED {msg.arbitration_id:08X}"
            else:
                if msg.is_remote_frame:
                    sendStr += f"CSR {msg.arbitration_id:03X}"
                else:
                    sendStr += f"CSD {msg.arbitration_id:03X}"

            sendStr += "".join([" %02X" % b for b in msg.data])

            self.write(sendStr)

    def shutdown(self):
        self.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2018-09-14
# @Filename: helpers.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import concurrent.futures
import enum
import logging
import warnings
from concurrent.futures import Executor
from contextlib import suppress
from functools import partial
from threading import Thread

from typing import TYPE_CHECKING, Callable, Generic, Optional, Type, TypeVar

from jaeger import log


if TYPE_CHECKING:
    from jaeger.actor import JaegerActor
    from jaeger.fps import FPS


__all__ = [
    "AsyncQueue",
    "StatusMixIn",
    "PollerList",
    "Poller",
    "AsyncioExecutor",
    "run_in_executor",
    "BaseBot",
]


class AsyncQueue(asyncio.Queue):
    """Provides an `asyncio.Queue` object with a watcher.

    Parameters
    ----------
    callback
        A function to call when a new item is received from the queue. It can
        be a coroutine.

    """

    def __init__(self, callback: Optional[Callable] = None):
        async def process_queue():
            """Waits for the next item and sends it to the cb function."""

            loop = asyncio.get_running_loop()
            while True:
                item = await self.get()
                if callback:
                    loop.call_soon_threadsafe(callback, item)

        super().__init__()

        self.watcher = asyncio.create_task(process_queue())


Status_co = TypeVar("Status_co", bound=enum.Enum)


class StatusMixIn(Generic[Status_co]):
    """A mixin that provides status tracking with callbacks.

    Provides a status property that executes a list of callbacks when
    the status changes.

    Parameters
    ----------
    maskbit_flags
        A class containing the available statuses as a series of maskbit
        flags. Usually as subclass of `enum.Flag`.
    initial_status
        The initial status.
    callback_func
        The function to call if the status changes.
    call_now
        Whether the callback function should be called when initialising.

    Attributes
    ----------
    callbacks
        A list of the callback functions to call.

    """

    def __init__(
        self,
        maskbit_flags: Type[Status_co],
        initial_status: Optional[Status_co] = None,
        callback_func: Optional[Callable] = None,
        call_now: bool = False,
    ):

        self._flags = maskbit_flags
        self.callbacks = []
        self._status: Optional[Status_co] = initial_status
        self.watcher = None

        if callback_func is not None:
            self.callbacks.append(callback_func)

        if call_now is True:
            self.do_callbacks()

    def add_callback(self, cb: Callable):
        """Adds a callback."""

        self.callbacks.append(cb)

    def remove_callback(self, cb: Callable):
        """Removes a callback."""

        self.callbacks.remove(cb)

    def do_callbacks(self):
        """Calls functions in ``callbacks``."""

        assert hasattr(
            self, "callbacks"
        ), "missing callbacks attribute. Did you call __init__()?"

        for func in self.callbacks:
            func()

    @property
    def status(self) -> Status_co:
        """Returns the status."""

        if self._status is None:
            raise ValueError("status not set.")

        return self._status

    @status.setter
    def status(self, value: Status_co):
        """Sets the status."""

        if value != self._status:
            self._status = self.flags(value)
            self.do_callbacks()
            if self.watcher is not None:
                self.watcher.set()

    @property
    def flags(self):
        """Gets the flags associated to this status."""

        return self._flags

    @flags.setter
    def flags(self, value: Type[Status_co]):
        """Sets the flags associated to this status."""

        self._flags = value
        self._status = None

    async def wait_for_status(
        self,
        value,
    ):
        """Awaits until the status matches ``value``."""

        if self.status == value:
            return

        self.watcher = asyncio.Event()

        while self.status != value:
            await self.watcher.wait()
            if self.watcher is not None:
                self.watcher.clear()

        self.watcher = None


class PollerList(list):
    """A list of `.Poller` to be managed jointly."""

    def __init__(self, pollers=[]):

        names = [poller.name for poller in pollers]
        assert len(names) == len(set(names)), "repeated names in poller list."

        list.__init__(self, pollers)

    @property
    def names(self):
        """List the poller names."""

        return [poller.name for poller in self]

    def append(self, poller):
        """Adds a poller."""

        assert isinstance(poller, Poller), "not a poller."

        names = [pp.name for pp in self]
        if poller.name in names:
            raise ValueError(
                f"a poller with name {poller.name} is already in the list."
            )

        list.append(self, poller)

    def __getattr__(self, name):
        """Gets a poller by its name."""

        for poller in self:
            if name == poller.name:
                return poller

        return list.__getitem__(self, name)

    def __getitem__(self, item):
        """Gets the poller by name."""

        if isinstance(item, str):
            return self.__getattr__(item)

        return list.__getitem__(self, item)

    async def set_delay(self, delay=None, immediate=False):
        """Sets the delay for all the pollers.

        Parameters
        ----------
        delay : float
            The delay between calls to the callback. If `None`, restores the
            original delay.
        immediate : bool
            If `True`, stops the currently running tasks and sets the
            new delay. Otherwise waits for the current tasks to complete.

        """

        delay_coros = [
            poller.set_delay(delay=delay, immediate=immediate) for poller in self
        ]
        await asyncio.gather(*delay_coros)

    def start(self, delay=None):
        """Starts all the pollers.

        Parameters
        ----------
        delay : float
            The delay between calls to the callback. If not specified,
            uses the default delays for each poller.

        """

        for poller in self:
            poller.start(delay=delay)

    async def stop(self):
        """Cancels all the poller."""

        stop_coros = [poller.stop() for poller in self]
        await asyncio.gather(*stop_coros)

    @property
    def running(self):
        """Returns `True` if at least one poller is running."""

        return any([poller.running for poller in self])


class Poller(object):
    """A task that runs a callback periodically.

    Parameters
    ----------
    name : str
        The name of the poller.
    callback : function or coroutine
        A function or coroutine to call periodically.
    delay : float
        Initial delay between calls to the callback.
    loop : event loop
        The event loop to which to attach the task.

    """

    def __init__(self, name, callback, delay=1, loop=None):

        self.name = name
        self.callback = callback

        self._orig_delay = delay
        self.delay = delay

        self.loop: asyncio.AbstractEventLoop = loop or asyncio.get_event_loop()

        # Create two tasks, one for the sleep timer and another for the poller
        # itself. We do this because we want to be able to cancell the sleep
        # coroutine if we are going to change the delay.
        self._sleep_task = None
        self._task = None

    async def poller(self):
        """The polling loop."""

        if self._task is None:
            raise RuntimeError("Task is not running.")

        while True:

            try:
                if asyncio.iscoroutinefunction(self.callback):
                    await self.callback()
                else:
                    self.callback()
            except Exception as ee:
                if ee.__class__ == asyncio.CancelledError:
                    raise
                if not self._task.cancelled:
                    self.loop.call_exception_handler(
                        {"message": "failed running callback", "exception": ee}
                    )

            self._sleep_task = self.loop.create_task(asyncio.sleep(self.delay))

            await self._sleep_task

    async def set_delay(self, delay=None, immediate=False):
        """Sets the delay for polling.

        Parameters
        ----------
        delay : float
            The delay between calls to the callback. If `None`, restores the
            original delay.
        immediate : bool
            If `True`, stops the currently running task and sets the
            new delay. Otherwise waits for the current task to complete.

        """

        # Only change delay if the difference is significant.
        if delay and abs(self.delay - delay) < 1e-6:
            return

        if not self.running:
            return

        if immediate:
            await self.stop()
            self.start(delay)
        else:
            self.delay = delay or self._orig_delay

    def start(self, delay=None):
        """Starts the poller.

        Parameters
        ----------
        delay : float
            The delay between calls to the callback. If not specified,
            restores the original delay used when the class was instantiated.

        """

        self.delay = delay or self._orig_delay

        if self.running:
            return

        self._task = self.loop.create_task(self.poller())

        return self

    async def stop(self):
        """Cancel the poller."""

        if self._task is None or not self.running:
            return

        self._task.cancel()

        with suppress(asyncio.CancelledError):
            if self._task is not None:
                await self._task

    async def call_now(self):
        """Calls the callback immediately."""

        restart = False
        delay = self.delay
        if self.running:
            await self.stop()
            restart = True

        if asyncio.iscoroutinefunction(self.callback):
            await self.loop.create_task(self.callback())
        else:
            self.callback()

        if restart:
            self.start(delay=delay)

    @property
    def running(self):
        """Returns `True` if the poller is running."""

        if self._task and not self._task.cancelled():
            return True

        return False


class AsyncioExecutor(Executor):
    """An executor to run coroutines from a normal function.

    Copied from http://bit.ly/2IYmqzN.

    To use, do ::

        with AsyncioExecutor() as executor:
            future = executor.submit(asyncio.sleep, 1)

    """

    def __init__(self):
        self._loop = asyncio.new_event_loop()
        self._thread = Thread(target=self._target)
        self._thread.start()

    def _target(self):
        asyncio.set_event_loop(self._loop)
        self._loop.run_forever()

    def submit(self, fn, *args, **kwargs):
        """Submit a coroutine to the executor."""

        coro = fn(*args, **kwargs)
        return asyncio.run_coroutine_threadsafe(coro, self._loop)

    def shutdown(self, wait=True):
        self._loop.call_soon_threadsafe(self._loop.stop)
        if wait:
            self._thread.join()


async def run_in_executor(fn, *args, catch_warnings=False, executor="thread", **kwargs):
    """Runs a function in an executor.

    In addition to streamlining the use of the executor, this function
    catches any warning issued during the execution and reissues them
    after the executor is done. This is important when using the
    actor log handler since inside the executor there is no loop that
    CLU can use to output the warnings.

    In general, note that the function must not try to do anything with
    the actor since they run on different loops.

    """

    fn = partial(fn, *args, **kwargs)

    if executor == "thread":
        executor = concurrent.futures.ThreadPoolExecutor
    elif executor == "process":
        executor = concurrent.futures.ProcessPoolExecutor
    else:
        raise ValueError("Invalid executor name.")

    if catch_warnings:
        with warnings.catch_warnings(record=True) as records:
            with executor() as pool:
                result = await asyncio.get_event_loop().run_in_executor(pool, fn)

        for ww in records:
            warnings.warn(ww.message, ww.category)

    else:
        with executor() as pool:
            result = await asyncio.get_running_loop().run_in_executor(pool, fn)

    return result


class BaseBot:
    """A class that monitors a subsystem."""

    def __init__(self, fps: FPS):

        self.fps = fps
        self.ieb = fps.ieb

        self.actor: JaegerActor | None = None

        self._task: asyncio.Task | None = None

    def set_actor(self, actor: JaegerActor):
        self.actor = actor

    async def start(self, delay: float | bool = False):
        """Stars the monitoring loop."""

        await self.stop()

        if delay is not False and delay > 0:
            await asyncio.sleep(delay)

        self._task = asyncio.create_task(self._loop())

    async def stop(self):
        """Stops the monitoring loop."""

        if self._task:
            with suppress(asyncio.CancelledError):
                self._task.cancel()
                await self._task

    async def restart(self):
        """Restarts the loop."""

        await self.stop()
        await self.start(delay=1)

    async def _loop(self):
        raise NotImplementedError()

    def notify(self, message: str | dict, level=logging.WARNING):
        """Logs a message and outputs it to the actor."""

        if isinstance(message, str):
            log.log(level, message)

            # No need to output to actor as well if this is a warning or above
            # because the ActorHandler already does it and we get duplicate messages.
            if level >= logging.WARNING:
                return

        if self.actor:
            message_code: str = "w"
            if level == logging.DEBUG:
                message_code = "d"
            elif level == logging.INFO:
                message_code = "i"
            elif level == logging.WARNING:
                message_code = "w"
            elif level == logging.ERROR:
                message_code = "e"

            if isinstance(message, str):
                message = {"text": message}

            self.actor.write(message_code, message=message)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-06-05
# @Filename: home.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio

from typing import TYPE_CHECKING

import click
import numpy

from jaeger import config
from jaeger.exceptions import JaegerError, TrajectoryError
from jaeger.kaiju import get_path_pair, get_robot_grid

from . import jaeger_parser


if TYPE_CHECKING:
    from jaeger import FPS

    from . import JaegerCommandType


__all__ = ["home"]


SAFE_BETA_MIN: float = 165.0
SAFE_BETA_MAX: float = 181.0

START_ANGLE: float = 5.0
COLLISION_BUFFER: float = 3.2

ALPHA1: float = 118
ALPHA2: float = 58

ALPHA_FOLDED, BETA_FOLDED = config["kaiju"]["lattice_position"]


@jaeger_parser.command()
@click.argument(
    "AXIS",
    type=click.Choice(["alpha", "beta"], case_sensitive=False),
)
@click.argument("POSITIONER_IDS", type=int, nargs=-1, required=False)
@click.option(
    "--start-angle",
    type=float,
    default=START_ANGLE,
    show_default=True,
    help="Angle from which to start the calibration.",
)
@click.option(
    "--skip-phase-1",
    is_flag=True,
    help="Do not run phase 1.",
)
@click.option(
    "-d",
    "--dry-run",
    is_flag=True,
    help="Execute trajectories but do not home the robots.",
)
async def home(
    command: JaegerCommandType,
    fps: FPS,
    axis: str,
    positioner_ids: tuple[int, ...] | list[int] = (),
    start_angle: float = 5.0,
    skip_phase_1: bool = False,
    dry_run: bool = False,
):
    """Re-homes positioner datums in bulk."""

    axis = axis.lower()

    positioner_ids = list(positioner_ids)
    if positioner_ids == []:
        positioner_ids = [
            pos.positioner_id
            for pos in fps.positioners.values()
            if not pos.disabled and not pos.offline
        ]
    else:
        for positioner_id in positioner_ids:
            if positioner_id not in fps.positioners.keys():
                return command.fail(f"Unknown positioner {positioner_id}.")
            if fps[positioner_id].disabled or fps[positioner_id].offline:
                return command.fail(f"Cannot home disabled positioner {positioner_id}.")

    if axis == "alpha":

        result = await check_positions(
            fps,
            beta_min=SAFE_BETA_MIN,
            beta_max=SAFE_BETA_MAX,
            check_disabled=True,
        )
        if not result:
            return command.fail("Beta arms are not safely folded.")

        command.info(f"Moving alpha to {start_angle} degrees.")
        try:
            await fps.goto(
                {
                    pos.positioner_id: (start_angle, pos.beta)
                    for pos in fps.positioners.values()
                    if pos.positioner_id in positioner_ids and pos.beta is not None
                }
            )
        except TrajectoryError as err:
            return command.fail(f"Trajectory failed with error {err}.")

        if dry_run is False:
            command.info("Homing in alpha.")
            home_tasks = []
            for pid in positioner_ids:
                home_tasks.append(fps[pid].home(alpha=True, beta=False))
            await asyncio.gather(*home_tasks)
        else:
            command.warning("Skipping homing in dry run.")

        command.info("Homing complete. Reverting to folded.")
        try:
            await fps.goto(
                {
                    pos.positioner_id: (ALPHA_FOLDED, pos.beta)
                    for pos in fps.positioners.values()
                    if pos.positioner_id in positioner_ids and pos.beta is not None
                },
                go_cowboy=True,
            )
        except TrajectoryError as err:
            return command.fail(f"Trajectory failed with error {err}.")

    elif axis == "beta":

        if not (await fps.is_folded()):
            return command.fail("The FPS is not folded. Cannot home beta.")

        command.info("Creating paths for phase 1.")

        grid1 = get_robot_grid(fps, collision_buffer=COLLISION_BUFFER)
        for robot in grid1.robotDict.values():
            if fps[robot.id].disabled is False:
                robot.setAlphaBeta(ALPHA1, start_angle)

        phase_2_pids = []
        for robot in grid1.robotDict.values():
            if robot.isOffline:
                continue
            if grid1.isCollided(robot.id):
                command.debug(f"Deferring positioner {robot.id} to phase 2.")
                robot.setAlphaBeta(ALPHA1, 180)
                phase_2_pids.append(robot.id)

        # All robots that should be moved in phase 1, even if won't be homed.
        phase_1_pids = [pid for pid in positioner_ids if pid not in phase_2_pids]
        # Robots in phase 1 that will be homed.
        phase_1_home_pids = list(set(positioner_ids) & set(phase_1_pids))

        command.info("Creating paths for phase 2.")

        grid2 = get_robot_grid(fps, collision_buffer=COLLISION_BUFFER)
        for robot in grid2.robotDict.values():
            if robot.id in phase_1_home_pids and dry_run is False:
                robot.betaOffDeg = 0.0
            if robot.id in phase_2_pids:
                robot.setAlphaBeta(ALPHA2, start_angle)
            else:
                robot.setAlphaBeta(ALPHA2, 180)

        _, from_destination_2, failed, _ = get_path_pair(
            grid2,
            path_generation_mode="greedy",
        )
        if failed:
            return command.fail("Failed generating paths for beta homing phase 2.")

        if skip_phase_1:
            phase_1_home_pids = []

        if len(phase_1_home_pids) > 0:

            _, from_destination_1, failed, _ = get_path_pair(
                grid1,
                path_generation_mode="greedy",
            )
            if failed:
                return command.fail("Failed generating paths for beta homing phase 1.")

            try:
                await _home_beta_phase(
                    command,
                    fps,
                    phase_1_home_pids,
                    1,
                    start_angle,
                    from_destination_1,
                    dry_run=dry_run,
                )
            except JaegerError as err:
                return command.fail(f"Phase 1 homing failed with error: {err}")

        else:

            command.info("No selected positioners in phase 1. Skipping ")

        # Robots in phase 2 that will be homed.
        phase_2_home_pids = list(set(positioner_ids) & set(phase_2_pids))

        if len(phase_2_home_pids) > 0:

            try:
                await _home_beta_phase(
                    command,
                    fps,
                    phase_2_home_pids,
                    2,
                    start_angle,
                    from_destination_2,
                    dry_run=dry_run,
                    extra_zero_positioner_ids=phase_1_home_pids,
                )
            except JaegerError as err:
                return command.fail(f"Phase 2 homing failed with error: {err}")

        else:

            command.info("No selected positioners in phase 2. Skipping ")

    return command.finish("Homing complete.")


async def check_positions(
    fps: FPS,
    alpha_min: float | None = None,
    alpha_max: float | None = None,
    beta_min: float | None = None,
    beta_max: float | None = None,
    check_disabled: bool = True,
):
    """Checks that positioners are in the correct place."""

    await fps.update_position()
    positions = fps.get_positions(ignore_disabled=not check_disabled)

    alpha_pos = positions[:, 1]
    if alpha_min is not None and alpha_max is not None:
        if numpy.any(alpha_pos < alpha_min) or numpy.any(alpha_pos > alpha_max):
            return False

    beta_pos = positions[:, 2]
    if beta_min is not None and beta_max is not None:
        if numpy.any(beta_pos < beta_min) or numpy.any(beta_pos > beta_max):
            return False

    return True


async def _home_beta_phase(
    command: JaegerCommandType,
    fps: FPS,
    home_positioner_ids: list[int],
    phase: int,
    start_angle: float,
    from_destination: dict,
    dry_run: bool = False,
    extra_zero_positioner_ids: list[int] = [],
):

    command.info(f"Moving robots to phase {phase} initial position.")
    try:
        await fps.send_trajectory(from_destination, command=command)
    except TrajectoryError as err:
        raise JaegerError(f"Trajectory failed with error: {err}")

    command.info(f"Starting phase {phase} calibration.")

    command.debug(f"Homing positioners {home_positioner_ids}.")

    if dry_run is False:
        command.info("Homing in beta.")
        home_tasks = []
        for pid in home_positioner_ids:
            home_tasks.append(fps[pid].home(alpha=False, beta=True))
        await asyncio.gather(*home_tasks)
    else:
        command.warning("Skipping homing in dry run.")

    command.info(f"Homing for phase {phase} finished.")
    command.info("Reverting robots to initial angle.")

    try:
        await fps.goto(
            {
                pos.positioner_id: (pos.alpha, start_angle)
                for pos in fps.positioners.values()
                if pos.positioner_id in home_positioner_ids and pos.alpha is not None
            },
            go_cowboy=True,
        )
    except TrajectoryError as err:
        raise JaegerError(f"Trajectory failed with error {err}.")

    # Although it's probably ok, we don't want to simply apply the to_destination
    # trajectory because the offsets have changed and that may cause a collision.
    # Instead we create a new grid, manually set the beta offsets to zero, then
    # calculate and execute an unwind.

    command.info("Reverting to folded.")

    await fps.update_position()

    grid_unwind = get_robot_grid(fps, collision_buffer=COLLISION_BUFFER)
    for robot in grid_unwind.robotDict.values():
        if dry_run is False:
            if robot.id in home_positioner_ids or robot.id in extra_zero_positioner_ids:
                robot.betaOffDeg = 0.0
        if robot.isOffline:
            continue
        robot.setAlphaBeta(fps[robot.id].alpha, fps[robot.id].beta)
        robot.setDestinationAlphaBeta(ALPHA_FOLDED, BETA_FOLDED)

    # This is equivalent to unwind with force=True but using the customised
    # grid with zeroed offsets. We want to force-unwind to ensure that most of
    # the robots make it home and there are only a few deadlocked robots that
    # need manual intervention.
    to_destination, _, did_fail, deadlocks = get_path_pair(
        grid_unwind,
        path_generation_mode="greedy",
        ignore_did_fail=True,
        stop_if_deadlock=True,
    )

    if len(deadlocks) > 0:
        command.warning("Deadlocks found in unwind trajectory but will unwind.")

    try:
        await fps.send_trajectory(to_destination, command=command)
    except TrajectoryError as err:
        raise JaegerError(f"Trajectory failed with error: {err}")

    if len(deadlocks) > 0:
        command.warning(f"Robots that have been homed in beta: {home_positioner_ids}")
        raise JaegerError(
            f"Failed to unwind robots {deadlocks}. Remove the beta offsets for the "
            "homed robots in fps_calibrations, then try to manually unwind."
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-04-24
# @Filename: actor.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import json
import logging
import os
from tempfile import NamedTemporaryFile
from time import time

import clu
import clu.protocol
from clu.tools import ActorHandler

import jaeger
from jaeger import FPS, __version__, log
from jaeger.alerts import AlertsBot
from jaeger.chiller import ChillerBot
from jaeger.exceptions import JaegerError, JaegerUserWarning


__all__ = ["JaegerActor"]


def merge_json(base: str, custom: str | None, write_temporary_file=False):
    """Merges two JSON files. Optional writes the result as a temporary file."""

    if custom is None:
        return base

    if os.path.samefile(base, custom):
        return base
    else:
        base_json = json.loads(open(base, "r").read())
        custom_json = json.loads(open(custom, "r").read())
        base_json["properties"].update(custom_json["properties"])

        if write_temporary_file is False:
            return base_json
        else:
            tmpfile = NamedTemporaryFile(delete=False, mode="w+")
            json.dump(base_json, tmpfile)
            tmpfile.flush()
            return tmpfile.name


class JaegerActor(clu.LegacyActor):
    """The jaeger SDSS-style actor."""

    def __init__(self, fps: FPS, *args, observatory: str | None = None, **kwargs):

        jaeger.actor_instance = self

        self.fps = fps

        # This is mostly for the miniwoks. If the schema file is not the base
        # one, merge them.
        base = os.path.join(os.path.dirname(__file__), "..")

        base_schema = os.path.realpath(os.path.join(base, "etc/schema.json"))

        schema = kwargs.get("schema", None)
        c_schema = os.path.realpath(os.path.join(base, schema)) if schema else None

        kwargs["schema"] = merge_json(base_schema, c_schema, write_temporary_file=True)

        # Pass the FPS instance as the second argument to each parser
        # command (the first argument is always the actor command).
        self.parser_args = [fps]

        # Set the observatory where the actor is running.
        if observatory is None:
            try:
                self.observatory = os.environ["OBSERVATORY"]
            except KeyError:
                raise JaegerError("Observatory not passed and $OBSERVATORY is not set.")
        else:
            self.observatory = observatory

        super().__init__(*args, **kwargs)

        self.version = __version__

        # Add ActorHandler to log and to the warnings logger.
        self.actor_handler = ActorHandler(
            self,
            level=logging.WARNING,
            filter_warnings=[JaegerUserWarning],
        )
        log.addHandler(self.actor_handler)
        if log.warnings_logger:
            log.warnings_logger.addHandler(self.actor_handler)

        self._alive_task = asyncio.create_task(self._report_alive())
        self._status_watcher_task = asyncio.create_task(self._status_watcher())

        # Define alerts and chiller bots.
        self.alerts = AlertsBot(self.fps)

        chiller_config = self.config["chiller"].get("config", None)
        self.chiller = ChillerBot(self.fps) if chiller_config else None

    async def start(self, *args, **kwargs):
        """Starts the actor and the bots."""

        await super().start(*args, **kwargs)

        self.alerts.set_actor(self)
        await self.alerts.start()

        if self.chiller:
            self.chiller.set_actor(self)
            await self.chiller.start()

    async def stop(self):
        """Stops the actor and bots."""

        await self.alerts.stop()
        if self.chiller:
            await self.chiller.stop()

        return await super().stop()

    async def start_status_server(self, port, delay=1):
        """Starts a server that outputs the status as a JSON on a timer."""

        self.status_server = clu.protocol.TCPStreamPeriodicServer(
            self.host,
            port,
            periodic_callback=self._report_status_cb,
            sleep_time=delay,
        )

        await self.status_server.start()

        self.log.info(f"starting status server on {self.host}:{port}")

    async def _report_alive(self):
        """Outputs the ``alive_at`` keyword."""

        while True:
            self.write("d", {"alive_at": time()}, broadcast=True)
            await asyncio.sleep(60)

    async def _report_status_cb(self, transport):
        """Reports the status to the status server."""

        status = await self.fps.report_status()
        status_json = json.dumps(status)

        transport.write(status_json.encode() + "\n".encode())

        return status

    async def _status_watcher(self):
        """Listens to the status async generator."""

        async for status in self.fps.async_status():
            self.write("i", fps_status=f"0x{status.value:x}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-07-12
# @Filename: bus.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import abc

from .message import Message


class BusABC(object, metaclass=abc.ABCMeta):
    """A base CAN bus."""

    def __init__(self, *args, **kwargs):
        pass

    async def open(self, *args, **kwargs) -> bool:
        """Starts the bus.

        This method call the ``_open_internal`` method in the subclass bus, if
        present. It's meant mainly to initialise any process that needs to be
        run as a coroutine. Must return `True` if the connection was successful,
        `False` or an error otherwise.

        """

        return await self._open_internal(*args, **kwargs)

    async def _open_internal(self):
        return True

    @abc.abstractmethod
    async def get(self):
        """Receives messages from the bus."""

        pass

    @abc.abstractmethod
    def send(self, msg: Message, **kwargs):
        """Sends a message to the bus."""

        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2018-08-27
# @Filename: __init__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import enum

from typing import Dict, Type, Union


class TypesEnumMeta(enum.EnumMeta):
    """Metaclass to allow initialising an Enum from a string."""

    def __call__(cls: Type[enum.Enum], value: Union[str, int]):  # type: ignore

        if isinstance(value, str):
            for flag in cls:
                if flag.name.lower() == value.lower():
                    return cls(flag.value)

        return super().__call__(value)


class CommandID(enum.IntEnum, metaclass=TypesEnumMeta):
    """IDs associated with commands."""

    GET_ID = 1
    GET_FIRMWARE_VERSION = 2
    GET_STATUS = 3
    SEND_NEW_TRAJECTORY = 10
    SEND_TRAJECTORY_DATA = 11
    TRAJECTORY_DATA_END = 12
    SEND_TRAJECTORY_ABORT = 13
    START_TRAJECTORY = 14
    STOP_TRAJECTORY = 15
    COLLISION_DETECTED = 18
    GO_TO_DATUMS = 20
    GO_TO_DATUM_ALPHA = 21
    GO_TO_DATUM_BETA = 22
    START_DATUM_CALIBRATION = 23
    START_DATUM_CALIBRATION_ALPHA = 24
    START_DATUM_CALIBRATION_BETA = 25
    START_MOTOR_CALIBRATION = 26
    START_MOTOR_CALIBRATION_ALPHA = 26
    START_MOTOR_CALIBRATION_BETA = 27
    GO_TO_ABSOLUTE_POSITION = 30
    GO_TO_RELATIVE_POSITION = 31
    GET_ACTUAL_POSITION = 32
    SET_ACTUAL_POSITION = 33
    GET_OFFSETS = 34
    SET_OFFSETS = 35
    SET_SPEED = 40
    SET_CURRENT = 41
    GET_HALL_CALIB_ERROR = 45
    START_COGGING_CALIBRATION = 47
    START_COGGING_CALIBRATION_ALPHA = 48
    START_COGGING_CALIBRATION_BETA = 49
    SAVE_INTERNAL_CALIBRATION = 53
    GET_CURRENT = 56
    GET_ALPHA_HALL_CALIB = 104
    GET_BETA_HALL_CALIB = 105
    SET_INCREASE_COLLISION_MARGIN = 111
    SET_HOLDING_CURRENT = 112
    GET_HOLDING_CURRENT = 113
    HALL_ON = 116
    HALL_OFF = 117
    ALPHA_CLOSED_LOOP_COLLISION_DETECTION = 118
    ALPHA_CLOSED_LOOP_WITHOUT_COLLISION_DETECTION = 119
    ALPHA_OPEN_LOOP_COLLISION_DETECTION = 120
    ALPHA_OPEN_LOOP_WITHOUT_COLLISION_DETECTION = 121
    BETA_CLOSED_LOOP_COLLISION_DETECTION = 122
    BETA_CLOSED_LOOP_WITHOUT_COLLISION_DETECTION = 123
    BETA_OPEN_LOOP_COLLISION_DETECTION = 124
    BETA_OPEN_LOOP_WITHOUT_COLLISION_DETECTION = 125
    SWITCH_LED_ON = 126
    SWITCH_LED_OFF = 127
    SWITCH_ON_PRECISE_MOVE_ALPHA = 128
    SWITCH_OFF_PRECISE_MOVE_ALPHA = 129
    SWITCH_ON_PRECISE_MOVE_BETA = 130
    SWITCH_OFF_PRECISE_MOVE_BETA = 131
    GET_RAW_TEMPERATURE = 132
    GET_NUMBER_TRAJECTORIES = 139
    SET_NUMBER_TRAJECTORIES = 140
    START_FIRMWARE_UPGRADE = 200
    SEND_FIRMWARE_DATA = 201

    def get_command_class(self) -> Type[Command]:
        """Returns the class associated with this command."""

        if self in COMMAND_LIST:
            return COMMAND_LIST[self]

        raise ValueError("The command does not have an associated class.")


from .base import *
from .base import Command
from .bootloader import *
from .calibration import *
from .goto import *
from .status import *
from .trajectory import *


def all_subclasses(cls):
    """Recursive subclasses."""

    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in all_subclasses(c)]
    )


# Generate a dictionary of commands
COMMAND_LIST: Dict[int, Type[Command]] = {
    cclass.command_id: cclass
    for cclass in all_subclasses(Command)
    if cclass.command_id in CommandID
}

# Dynamically generate command classes for those commands for which we
# didn't write a class. These classes are identical to a generic Command
# but with a custom name.
for cid in list(CommandID):
    if cid not in COMMAND_LIST:
        CommandClass = type(
            cid.name.title().replace("_", ""),
            (Command,),
            {"command_id": cid, "broadcastable": False},
        )
        COMMAND_LIST[cid] = CommandClass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-10-13
# @Filename: design.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from dataclasses import dataclass

import peewee

from coordio.defaults import calibration
from sdssdb.peewee.sdss5db import targetdb

from jaeger.utils.database import connect_database
from jaeger.utils.helpers import run_in_executor

from .configuration import Configuration


__all__ = ["Design"]


@dataclass
class FieldData:
    """Field data for design."""

    field_id: int
    rs_run: str
    observatory: str
    racen: float
    deccen: float
    position_angle: float


class Design:
    """Loads and represents a targetdb design.

    Parameters
    ----------
    design_id
        The ID of the design to load.
    load_configuration
        Create a `.Configuration` attached to this design.
    epoch
        The JD epoch for which to calculate the configuration coordinates. If
        `None`, uses the current time.
    scale
        Focal plane scale factor to apply. Defaults to coordio's internal value.

    """

    def __init__(
        self,
        design_id: int,
        load_configuration: bool = True,
        epoch: float | None = None,
        scale: float | None = None,
    ):

        if calibration.wokCoords is None:
            raise RuntimeError("Cannot retrieve wok calibration. Is $WOKCALIB_DIR set?")

        self.design_id = design_id

        if connect_database(targetdb.database) is False:
            raise RuntimeError("Cannot connect to database.")

        try:
            self.design = targetdb.Design.get(design_id=design_id)
        except peewee.DoesNotExist:
            raise ValueError(f"design_id {design_id} does not exist in the database.")

        self.field = FieldData(
            field_id=self.design.field.field_id,
            rs_run=self.design.field.version.plan if self.design else "NA",
            observatory=self.design.field.observatory.label,
            racen=self.design.field.racen,
            deccen=self.design.field.deccen,
            position_angle=self.design.field.position_angle,
        )
        self.target_data: dict[str, dict] = self.get_target_data()

        self.configuration: Configuration
        if load_configuration:
            self.configuration = Configuration(self, epoch=epoch, scale=scale)

    def get_target_data(self) -> dict[str, dict]:
        """Retrieves target data as a dictionary."""

        # TODO: this is all synchronous which is probably ok because this
        # query should run in < 1s, but at some point maybe we can change
        # this to use async-peewee and aiopg.

        if connect_database(targetdb.database) is False:
            raise RuntimeError("Database is not connected.")

        target_data = (
            targetdb.Design.select(
                targetdb.Assignment.pk.alias("assignment_pk"),
                targetdb.CartonToTarget.pk.alias("carton_to_target_pk"),
                targetdb.CartonToTarget.lambda_eff,
                targetdb.CartonToTarget.delta_ra,
                targetdb.CartonToTarget.delta_dec,
                targetdb.Target,
                targetdb.Magnitude,
                targetdb.Hole.holeid,
                targetdb.Instrument.label.alias("fibre_type"),
                targetdb.Cadence.label.alias("cadence"),
                targetdb.Carton.carton,
                targetdb.Category.label.alias("category"),
                targetdb.Carton.program,
            )
            .join(targetdb.Assignment)
            .join(targetdb.CartonToTarget)
            .join(targetdb.Target)
            .switch(targetdb.CartonToTarget)
            .join(targetdb.Carton)
            .join(targetdb.Category, peewee.JOIN.LEFT_OUTER)
            .switch(targetdb.CartonToTarget)
            .join(targetdb.Cadence, peewee.JOIN.LEFT_OUTER)
            .switch(targetdb.CartonToTarget)
            .join(targetdb.Magnitude, peewee.JOIN.LEFT_OUTER)
            .switch(targetdb.Assignment)
            .join(targetdb.Hole)
            .switch(targetdb.Assignment)
            .join(targetdb.Instrument)
            .where(targetdb.Design.design_id == self.design_id)
            .dicts()
        )

        return {data["holeid"]: data for data in target_data}

    @classmethod
    def check_design(cls, design_id: int, site: str):
        """Checks if a design exists and is for the current observatory."""

        if connect_database(targetdb.database) is False:
            raise RuntimeError("Database is not connected.")

        exists = (
            targetdb.Design.select()
            .where(targetdb.Design.design_id == design_id)
            .exists()
        )
        if exists is False:
            return False

        observatory = (
            targetdb.Design.select(targetdb.Observatory.label)
            .join(targetdb.DesignToField)
            .join(targetdb.Field)
            .join(targetdb.Observatory)
            .where(targetdb.Design.design_id == design_id)
            .limit(1)
            .scalar()
        )

        return site == observatory

    @classmethod
    async def create_async(
        cls,
        design_id: int,
        epoch: float | None = None,
        scale: float | None = None,
    ):
        """Returns a design while creating the configuration in an executor."""

        self = cls(design_id, load_configuration=False)

        configuration = await run_in_executor(
            Configuration,
            self,
            epoch=epoch,
            scale=scale,
        )
        self.configuration = configuration

        return self

    def __repr__(self):
        return f"<Design (design_id={self.design_id})>"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-11-01
# @Filename: fvc.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import logging
import os
import pathlib
import warnings

from typing import TYPE_CHECKING, Optional

import numpy
import pandas
from astropy.io import fits
from astropy.table import Table

from clu.command import Command
from clu.legacy.tron import TronConnection
from coordio.defaults import calibration

from jaeger import config, log
from jaeger.exceptions import FVCError, JaegerUserWarning, TrajectoryError
from jaeger.fps import FPS
from jaeger.ieb import IEB
from jaeger.kaiju import get_path_pair_in_executor, get_robot_grid
from jaeger.plotting import plot_fvc_distances
from jaeger.target import Configuration, Design
from jaeger.target.tools import read_confSummary, wok_to_positioner
from jaeger.utils import run_in_executor


if TYPE_CHECKING:
    from coordio.transforms import FVCTransformAPO, FVCTransformLCO

    from jaeger.actor import JaegerActor
    from jaeger.target.configuration import BaseConfiguration


__all__ = ["FVC"]


FVC_CONFIG = config["fvc"]


def get_transform(observatory: str):
    """Returns the correct coordio FVC transform class for the observatory."""

    if observatory.upper() == "APO":
        from coordio.transforms import FVCTransformAPO

        return FVCTransformAPO

    elif observatory.upper() == "LCO":
        from coordio.transforms import FVCTransformLCO

        return FVCTransformLCO

    else:
        raise ValueError(f"Invalid observatory {observatory}.")


class FVC:
    """Focal View Camera class."""

    configuration: Optional[BaseConfiguration]
    fibre_data: Optional[pandas.DataFrame]
    measurements: Optional[pandas.DataFrame]
    centroids: Optional[pandas.DataFrame]
    offsets: Optional[pandas.DataFrame]
    fvc_transform: Optional[FVCTransformAPO | FVCTransformLCO]

    image_path: Optional[str]
    proc_image_path: Optional[str]
    raw_hdu: Optional[fits.ImageHDU]
    proc_hdu: Optional[fits.ImageHDU]

    centroid_method: str | None
    fitrms: float
    k: float

    def __init__(self, site: str, command: Optional[Command[JaegerActor]] = None):

        if len(calibration.positionerTable) == 0:
            warnings.warn(
                "FPS calibrations not loaded or the array is empty.",
                JaegerUserWarning,
            )

        self.site = site
        self.correction_applied: bool = False

        self.command = command
        self.fps = FPS.get_instance()

        # To be updated manually by the actor command.
        self.iteration: int = 1

        self.reset()

    def reset(self):
        """Resets the instance."""

        self.configuration = None
        self.fvc_transform: FVCTransformAPO | FVCTransformLCO | None = None
        self.fibre_data = None
        self.centroids = None
        self.offsets = None

        self.image_path = None
        self.proc_image_path = None
        self.raw_hdu = None
        self.proc_hdu = None

        self.k = 1
        self.fitrms = -9.99
        self.perc_90 = -9.99
        self.fvc_percent_reached = -9.99
        self.centroid_method = config["fvc"]["centroid_method"]

    def set_command(self, command: Command[JaegerActor]):
        """Sets the command."""

        if command.status.is_done:
            raise FVCError("Command is done.")

        self.command = command

    def log(
        self,
        msg: str,
        level: logging._Level = logging.INFO,
        to_log: bool = True,
        to_command: bool = True,
    ):
        """Logs a message, including to the command if present."""

        level = int(level)
        msg = "[FVC]: " + msg

        if log and to_log:
            log.log(level, msg)

        if self.command and to_command:
            if level == logging.DEBUG:
                self.command.debug(msg)
            elif level == logging.INFO:
                self.command.info(msg)
            elif level == logging.WARNING:
                self.command.warning(msg)
            elif level == logging.ERROR:
                self.command.error(msg)

    async def expose(
        self,
        exposure_time: float = 5.0,
        stack: int = 1,
        use_tron_fallback=True,
    ) -> pathlib.Path:  # pragma: no cover
        """Takes an exposure with the FVC and blocks until the exposure is complete.

        Returns the path to the new image. If ``use_tron_fallback=True`` and the
        command has not been set, creates a Tron client to command the FVC.

        """

        if self.command is None:
            if use_tron_fallback is False:
                raise FVCError("Command must be set.")

        else:
            if self.command.status.is_done:
                raise FVCError("Command is done.")

        self.log(
            f"Taking {stack} x {exposure_time} seconds FVC exposure.",
            to_command=False,
        )

        tron = None

        if stack <= 1:
            cmd_str = f"talk -c fvc expose {exposure_time}"
        else:
            cmd_str = f"talk -c fvc expose --stack {stack} {exposure_time}"

        if self.command:
            expose_command = self.command.send_command("fliswarm", cmd_str)
        else:
            tron = TronConnection(
                "jaeger.jaeger",
                config["actor"]["tron_host"],
                config["actor"]["tron_port"],
            )
            await tron.start()
            expose_command = tron.send_command("fliswarm", cmd_str)

        assert isinstance(expose_command, Command)
        await expose_command

        if tron:
            tron.stop()

        if expose_command.status.did_fail:
            raise FVCError("The FVC exposure failed.")

        try:
            filename = expose_command.replies.get("filename")[-1]
            self.log(f"FVC raw image is {filename}.", to_command=False)
            return pathlib.Path(filename)
        except KeyError:
            raise FVCError("The exposure succeeded but did not output the filename.")

    def process_fvc_image(
        self,
        path: pathlib.Path | str,
        positioner_coords: dict,
        configuration: Optional[BaseConfiguration] = None,
        fibre_data: Optional[pandas.DataFrame] = None,
        fibre_type: str = "Metrology",
        centroid_method: str | None = None,
        use_new_invkin: bool = True,
        plot: bool | str = False,
        outdir: str | None = None,
    ) -> tuple[fits.ImageHDU, pandas.DataFrame, pandas.DataFrame | None]:
        """Processes a raw FVC image.

        Parameters
        ----------
        path
            The path to the raw FVC image.
        positioner_coords
            A dictionary of positioner ID to ``(alpha, beta)`` with the current
            positions of the robots.
        configuration
            A configuration object to use for processing. If `None`, defaults to the
            current `.FPS` loaded configuration.
        fibre_data
            A Pandas data frame with the expected coordinates of the targets. It
            is expected the data frame will have columns ``positioner_id``,
            ``hole_id``, ``fibre_type``, ``xwok``, and ``ywok``. This frame is
            appended to the processed image. Normally this parameters is left
            empty and the fibre table from the configuration loaded into the FPS
            instace is used.
        fibre_type
            The ``fibre_type`` rows in ``fibre_data`` to use. Defaults to
            ``fibre_type='Metrology'``.
        centroid_method
            The centroid method to use, one of ``"nudge"``, ``"sep"``, ``"winpos"``,
            or ``"simple"``. Defaults to ``"nudge"``.
        use_new_invkin
            Use new inverse kinnematic to calculate alpha/beta.
        plot
            Whether to save additional debugging plots along with the processed image.
            If ``plot`` is a string, it will be used as the directory to which to
            save the plots.

        Returns
        -------
        result
            A tuple with the read raw image HDU (with columns flipped) as the first
            argument and the expected coordinates, as a data frame, as the second.
            The data frame is the same as the input target coordinates but with the
            columns ``xwok_measured`` and ``ywok_measured`` appended.

        """

        # Reset the instance
        self.reset()

        centroid_method = centroid_method or self.centroid_method

        path = str(path)
        if not os.path.exists(path):
            raise FVCError(f"FVC image {path} does not exist.")

        self.configuration = configuration or self.fps.configuration

        if fibre_data is None:
            if configuration is None:
                raise FVCError("No fibre data and no configuration has been loaded.")
            fibre_data = configuration.assignment_data.fibre_table

        fdata = fibre_data.copy().reset_index().set_index("positioner_id")

        # Set fibre_data initially so that if FVCTransform fails we have something.
        self.fibre_data = fdata.copy().reset_index()
        self.fibre_data.set_index(["hole_id", "fibre_type"], inplace=True)

        self.log(f"Processing raw image {path}")

        dirname, base = os.path.split(path)
        dirname = outdir or dirname

        proc_path_root = os.path.join(dirname, "proc-" + base[0 : base.find(".fit")])

        if plot is True:
            plot_path_root = proc_path_root
        elif plot is False or plot is None:
            plot_path_root = None
        elif isinstance(plot, str):
            plot_path_root = os.path.join(plot, "proc-" + base[0 : base.find(".fit")])
        else:
            plot_path_root = ""

        hdus = fits.open(path)

        self.image_path = path
        self.raw_hdu = hdus[1].copy()
        self.proc_hdu = hdus[1]

        # Invert columns at APO.
        if self.fps.observatory == "APO":
            hdus[1].data = hdus[1].data[:, ::-1]

        image_data = hdus[1].data

        self.log(f"Max counts in image: {numpy.max(image_data)}", level=logging.INFO)

        # Get the rotator angle so that we can derotate the centroids to the
        # x/ywok-aligned configuration.
        rotpos = hdus[1].header.get("IPA", 135.4)
        if rotpos is None:
            raise FVCError("IPA keyword not found in the header.")
        rotpos = float(rotpos) % 360.0

        positioner_df = pandas.DataFrame.from_dict(
            positioner_coords,
            columns=["alphaReport", "betaReport"],
            orient="index",
        )
        positioner_df.index.set_names(["positionerID"], inplace=True)

        FVCTransform = get_transform(self.fps.observatory)
        self.log(f"Using FVC transform class {FVCTransform.__name__!r}.")

        fvc_transform = FVCTransform(
            image_data,
            positioner_df,
            rotpos,
            plotPathPrefix=plot_path_root,
        )

        self.centroids = fvc_transform.extractCentroids()
        fvc_transform.fit(centType=centroid_method, newInvKin=use_new_invkin)
        self.centroid_method = fvc_transform.centType

        self.log(f"Centroid method: {self.centroid_method}.")

        if self.command:
            self.command.info(fvc_centroid_method=self.centroid_method)

        assert fvc_transform.positionerTableMeas is not None

        measured = fvc_transform.positionerTableMeas.copy().set_index("positionerID")

        n_dubious = measured.wokErrWarn.sum()
        if n_dubious > 0:
            self.log(
                f"Found {n_dubious} positioners with dubious centroid matches.",
                level=logging.WARNING,
            )

        metrology_data = fdata.copy().reset_index()
        metrology_data = metrology_data.loc[metrology_data.fibre_type == fibre_type]

        # Create a column to mark positioners with dubious matches.
        fdata.loc[:, "dubious"] = 0
        dubious_pid = measured.loc[measured.wokErrWarn].index.values
        fdata.loc[dubious_pid, "dubious"] = 1

        wok_measured = measured.loc[
            metrology_data.positioner_id, ["xWokMeasMetrology", "yWokMeasMetrology"]
        ]
        fdata.loc[
            fdata.fibre_type == fibre_type, ["xwok_measured", "ywok_measured"]
        ] = wok_measured.values

        fdata = fdata.reset_index().set_index("fibre_type")

        # Only use online, assigned robots for final RMS. First get groups of fibres
        # with an assigned robot, that are not offline or dubious.
        if fdata.assigned.sum() > 0:
            assigned = fdata.groupby("positioner_id").filter(
                lambda g: g.assigned.any()
                & (g.offline == 0).all()
                & (g.dubious == 0).all()
                # & (g.on_target == 1).any()
            )
        else:
            self.log("No assigned fibres found. Using all matched fibres.")
            assigned = fdata.groupby("positioner_id").filter(
                lambda g: (g.offline == 0).all()
                & (g.dubious == 0).all()
                # & (g.on_target == 1).any()
            )

        # Now get the metrology fibre from those groups.
        assigned = assigned.loc[fibre_type]

        # Calculate RMS from assigned fibres.
        dx = assigned.xwok - assigned.xwok_measured
        dy = assigned.ywok - assigned.ywok_measured
        self.fitrms = numpy.round(numpy.sqrt(numpy.mean(dx**2 + dy**2)), 5)
        self.log(f"RMS full fit {self.fitrms * 1000:.3f} um.")

        # Also calculate 90% percentile and percentage of targets below threshold.
        distance = numpy.sqrt(dx**2 + dy**2)

        self.perc_90 = numpy.round(numpy.percentile(distance, 90), 4)

        n_reached = numpy.sum(distance <= (config["fvc"]["target_distance"] / 1000))
        self.fvc_percent_reached = numpy.round(n_reached / len(dx) * 100, 1)

        # FITSRMS is the RMS of measured - expected for assigned, non-disabled
        # robots. This is different from FVC_RMS reported by
        # FVCTransformAPO.getMetadata() that is measured - reported for all
        # positioners.
        hdus[1].header["FVCITER"] = (self.iteration, "FVC iteration")
        hdus[1].header["FITRMS"] = (self.fitrms * 1000, "RMS full fit [um]")
        hdus[1].header["PERC90"] = (self.perc_90 * 1000, "90% percentile [um]")
        hdus[1].header["FVCREACH"] = (
            self.fvc_percent_reached,
            "Targets that have reached their goal [%]",
        )

        fdata.reset_index(inplace=True)
        fdata.set_index(["hole_id", "fibre_type"], inplace=True)

        self.fibre_data = fdata
        self.fvc_transform = fvc_transform

        self.log(f"Finished processing {path}", level=logging.DEBUG)

        return (self.proc_hdu, self.fibre_data, self.centroids)

    def calculate_offsets(
        self,
        reported_positions: numpy.ndarray,
        fibre_data: Optional[pandas.DataFrame] = None,
        k: Optional[float] = None,
        max_correction: Optional[float] = None,
    ) -> pandas.DataFrame:
        """Determines the offset to apply to the currently reported positions.

        Measured wok positions from the fibre data are converted to positioner
        coordinates. An alpha/beta offset is calculated with respect to the
        expected positions. The offsets is then applied to the current positions
        as self-reported by the positioners. Optionally, the offset can be
        adjusted using a PID loop.

        Parameters
        ----------
        reported_positions
            Reported positions for the positioners as a numpy array. Usually
            the output of `.FPS.get_positions`.
        fibre_data
            The fibre data table. Only the metrology entries are used. Must
            have the ``xwok_measured`` and ``ywok_measured`` column populated.
            If `None`, uses the data frame calculated when `.process_fvc_image`
            last run.
        k
            The fraction of the correction to apply.
        max_correction
            The maximum offset allowed per robot and arm, in degrees. Corrections
            larger than ``max_offset`` are clipped.

        Returns
        -------
        new_positions
            The new alpha and beta positions as a Pandas dataframe indexed by
            positions ID. If `None`, uses the value ``fvc.k`` from the configuration.

        """

        self.log("Calculating offset from FVC image and model fit.")

        site = config["observatory"]

        self.k = k or config["fvc"]["k"]
        max_offset: float = max_correction or config["fvc"]["max_correction"]

        if fibre_data is None and self.fibre_data is None:
            raise FVCError("No fibre data passed or stored in the instance.")

        if fibre_data is None:
            fibre_data = self.fibre_data
            assert fibre_data is not None

        fibre_data = fibre_data.copy().reset_index()
        met: pandas.DataFrame = fibre_data.loc[fibre_data.fibre_type == "Metrology"]

        # TODO: deal with missing data
        if (met.loc[:, ["xwok_measured", "ywok_measured"]] == -999.0).any().any():
            raise FVCError("Some metrology fibres have not been measured.")

        # Calculate alpha/beta from measured wok coordinates.
        _measured = []
        first = True
        for _, row in met.iterrows():
            (alpha_measured, beta_measured), _ = wok_to_positioner(
                row.hole_id,
                site,
                "Metrology",
                row.xwok_measured,
                row.ywok_measured,
            )

            if "alpha" in row:
                alpha_expected = row.alpha
                beta_expected = row.beta
            else:
                if first:
                    self.log(
                        "Fibre data does not include the expected alpha/beta "
                        "positions. Using reported alpha/beta.",
                        logging.WARNING,
                    )
                prow = reported_positions[reported_positions[:, 0] == row.positioner_id]
                alpha_expected = prow[0][1]
                beta_expected = prow[0][2]

            # If beta >= 180, we would need a left handed configuration. For now we
            # invalidate these values.
            if beta_expected >= 180.0:
                alpha_measured = numpy.nan
                beta_measured = numpy.nan

            xwok_distance = row.xwok_measured - row.xwok
            ywok_distance = row.ywok_measured - row.ywok

            _measured.append(
                (
                    row.hole_id,
                    row.positioner_id,
                    xwok_distance,
                    ywok_distance,
                    alpha_expected,
                    beta_expected,
                    alpha_measured,
                    beta_measured,
                )
            )

            first = False

        measured = pandas.DataFrame(
            _measured,
            columns=[
                "hole_id",
                "positioner_id",
                "xwok_distance",
                "ywok_distance",
                "alpha_expected",
                "beta_expected",
                "alpha_measured",
                "beta_measured",
            ],
        )
        measured.set_index("positioner_id", inplace=True)

        # Merge the reported positions.
        reported = pandas.DataFrame(
            reported_positions,
            columns=["positioner_id", "alpha_reported", "beta_reported"],
        )
        reported.positioner_id = reported.positioner_id.astype("int32")
        reported.set_index("positioner_id", inplace=True)

        offsets = pandas.concat([measured, reported], axis=1)

        # If there are measured alpha/beta that are NaN, replace those with the
        # previous value.
        offsets["transformation_valid"] = 1
        pos_na = offsets["alpha_measured"].isna()
        if pos_na.sum() > 0:
            self.log(
                "Failed to calculate corrected positioner coordinates for "
                f"{pos_na.sum()} positioners.",
                level=logging.WARNING,
            )

            # For now set these values to the expected because we'll use them to
            # calculate the offset (which will be zero for invalid conversions).
            expected = offsets.loc[pos_na, ["alpha_expected", "beta_expected"]]
            offsets.loc[pos_na, ["alpha_measured", "beta_measured"]] = expected
            offsets.loc[pos_na, "transformation_valid"] = 0

        # Calculate offset between expected and measured.
        alpha_offset = offsets["alpha_expected"] - offsets["alpha_measured"]
        beta_offset = offsets["beta_expected"] - offsets["beta_measured"]

        offsets["alpha_offset"] = alpha_offset
        offsets["beta_offset"] = beta_offset

        # Clip very large offsets and apply a proportional term.
        alpha_offset_c = numpy.clip(self.k * alpha_offset, -max_offset, max_offset)
        beta_offset_c = numpy.clip(self.k * beta_offset, -max_offset, max_offset)

        offsets["alpha_offset_corrected"] = alpha_offset_c
        offsets["beta_offset_corrected"] = beta_offset_c

        # Calculate new alpha and beta by applying the offset to the reported
        # positions (i.e., the ones the positioners believe they are at).
        alpha_new = offsets["alpha_reported"] + offsets["alpha_offset_corrected"]
        beta_new = offsets["beta_reported"] + offsets["beta_offset_corrected"]

        # Get new positions that are out of range and clip them.
        alpha_oor = (alpha_new < 0.0) | (alpha_new > 360.0)
        beta_oor = (beta_new < 0.0) | (beta_new > 180.0)

        alpha_new.loc[alpha_oor] = numpy.clip(alpha_new.loc[alpha_oor], 0, 360)
        beta_new.loc[beta_oor] = numpy.clip(beta_new.loc[beta_oor], 0, 180)

        # For the values out of range, recompute the corrected offsets.
        alpha_corr = alpha_new.loc[alpha_oor] - offsets.loc[alpha_oor, "alpha_reported"]
        offsets.loc[alpha_oor, "alpha_offset_corrected"] = alpha_corr
        beta_corr = beta_new.loc[beta_oor] - offsets.loc[beta_oor, "beta_reported"]
        offsets.loc[beta_oor, "beta_offset_corrected"] = beta_corr

        # Final check. If alpha/beta_new are NaNs, replace with reported values.
        alpha_new[numpy.isnan(alpha_new)] = offsets.loc[
            numpy.isnan(alpha_new), "alpha_reported"
        ]
        beta_new[numpy.isnan(beta_new)] = offsets.loc[
            numpy.isnan(beta_new), "beta_reported"
        ]

        # Save new positions.
        offsets["alpha_new"] = alpha_new
        offsets["beta_new"] = beta_new

        # Set the invalid alpha/beta_measured back to NaN.
        invalid = offsets["transformation_valid"] == 0
        offsets.loc[
            invalid,
            [
                "alpha_measured",
                "beta_measured",
                "alpha_offset",
                "beta_offset",
                "alpha_offset_corrected",
                "beta_offset_corrected",
            ],
        ] = numpy.nan

        self.offsets = offsets

        self.log("Finished calculating offsets.", level=logging.DEBUG)

        return offsets

    async def write_proc_image(
        self,
        new_filename: Optional[str | pathlib.Path] = None,
    ) -> fits.HDUList:  # pragma: no cover
        """Writes the processed image along with additional table data.

        If ``new_filename`` is not passed, defaults to adding the prefix ``proc-``
        to the last processed image file path.

        """

        if self.image_path is None or self.proc_hdu is None:
            raise FVCError(
                "No current image. Take and process "
                "an image before callin write_proc_image()."
            )

        if new_filename is None:
            image_path = pathlib.Path(self.image_path)
            new_filename = image_path.with_name("proc-" + image_path.name)

        if self.fibre_data is None or self.raw_hdu is None or self.proc_hdu is None:
            raise FVCError("Need to run process_fvc_image before writing the image.")

        proc_hdus = fits.HDUList([fits.PrimaryHDU(), self.proc_hdu])

        if self.configuration:
            proc_hdus[1].header["CONFIGID"] = self.configuration.configuration_id
        else:
            proc_hdus[1].header["CONFIGID"] = -999.0

        proc_hdus[1].header["CAPPLIED"] = self.correction_applied

        positionerTable = calibration.positionerTable
        wokCoords = calibration.wokCoords
        fiducialCoords = calibration.fiducialCoords

        dfs = [
            ("POSITIONERTABLE", positionerTable.reset_index()),
            ("WOKCOORDS", wokCoords.reset_index()),
            ("FIDUCIALCOORDS", fiducialCoords.reset_index()),
        ]

        if self.fvc_transform is not None:
            if self.fvc_transform.positionerTableMeas is not None:
                pos_table_meas = self.fvc_transform.positionerTableMeas
                pos_table_meas = pos_table_meas.drop(columns=["index"])
                dfs.append(("POSITIONERTABLEMEAS", pos_table_meas))
            if self.fvc_transform.fiducialCoordsMeas is not None:
                fid_coords_meas = self.fvc_transform.fiducialCoordsMeas
                fid_coords_meas = fid_coords_meas.drop(columns=["index"])
                dfs.append(("FIDUCIALCOORDSMEAS", fid_coords_meas))

        for name, df in dfs:
            rec = Table.from_pandas(df).as_array()
            table = fits.BinTableHDU(rec, name=name)
            proc_hdus.append(table)

        fibre_data = self.fibre_data.copy()
        fibre_data.reset_index(inplace=True)
        fibre_data.sort_values("positioner_id", inplace=True)
        fibre_data_rec = Table.from_pandas(fibre_data).as_array()
        proc_hdus.append(fits.BinTableHDU(fibre_data_rec, name="FIBERDATA"))

        # Add IEB information
        ieb_keys = config["fvc"]["ieb_keys"]
        ieb_data = {key: -999.0 for key in ieb_keys}

        for key in ieb_keys:
            device_name = ieb_keys[key]
            if self.fps and isinstance(self.fps.ieb, IEB):
                try:
                    device = self.fps.ieb.get_device(device_name)
                    ieb_data[key] = (await device.read())[0]
                except Exception as err:
                    self.log(f"Failed getting IEB information: {err}", logging.WARNING)
                    break

        for key, val in ieb_data.items():
            proc_hdus[1].header[key] = val

        # Add header keywords from coordio.FVCTransfromAPO.
        if self.fvc_transform:
            try:
                proc_hdus[1].header.extend(self.fvc_transform.getMetadata())
            except Exception as err:
                self.log(f"Cannot get FVCTransform metadata: {err}", logging.WARNING)

        posangles = None
        if self.fps:
            await self.fps.update_position()
            positions = self.fps.get_positions()
            current_positions = pandas.DataFrame(
                {
                    "positionerID": positions[:, 0].astype(int),
                    "alphaReport": positions[:, 1],
                    "betaReport": positions[:, 2],
                }
            )

            current_positions["cmdAlpha"] = numpy.nan
            current_positions["cmdBeta"] = numpy.nan
            current_positions["startAlpha"] = numpy.nan
            current_positions["startBeta"] = numpy.nan

            if self.configuration:
                robot_grid = self.configuration.robot_grid

                _cmd_alpha = []
                _cmd_beta = []
                _start_alpha = []
                _start_beta = []

                if len(list(robot_grid.robotDict.values())[0].alphaPath) > 0:
                    for pid in current_positions.positionerID:
                        robot = robot_grid.robotDict[pid]
                        _cmd_alpha.append(robot.alphaPath[0][1])
                        _cmd_beta.append(robot.betaPath[0][1])
                        _start_alpha.append(robot.alphaPath[-1][1])
                        _start_beta.append(robot.betaPath[-1][1])

                    current_positions["cmdAlpha"] = _cmd_alpha
                    current_positions["cmdBeta"] = _cmd_beta
                    current_positions["startAlpha"] = _start_alpha
                    current_positions["startBeta"] = _start_beta

            posangles = Table.from_pandas(current_positions).as_array()
        proc_hdus.append(fits.BinTableHDU(posangles, name="POSANGLES"))

        if self.centroids is not None:
            rec = Table.from_pandas(self.centroids).as_array()
        else:
            rec = None
        proc_hdus.append(fits.BinTableHDU(rec, name="CENTROIDS"))

        offsets = None
        if self.offsets is not None:
            offsets = Table.from_pandas(self.offsets.reset_index()).as_array()
        proc_hdus.append(fits.BinTableHDU(offsets, name="OFFSETS"))

        await run_in_executor(proc_hdus.writeto, new_filename, checksum=True)

        self.log(f"Processed HDU written to {new_filename}")
        self.proc_image_path = os.path.abspath(new_filename)

        return proc_hdus

    async def apply_correction(
        self,
        offsets: Optional[pandas.DataFrame] = None,
    ):  # pragma: no cover
        """Applies the offsets. Fails if the trajectory is collided or deadlock."""

        if self.fps.locked:
            raise FVCError("The FPS is locked. Cannot apply corrections.")

        self.log("Preparing correction trajectory.")

        if self.offsets is None and offsets is None:
            raise FVCError("Offsets not set or passed. Cannot apply correction.")

        if offsets is None:
            offsets = self.offsets

        assert offsets is not None
        await self.fps.update_position()

        target_distance = config["fvc"]["target_distance"]

        # Setup robot grid.
        grid = get_robot_grid(self.fps)
        for robot in grid.robotDict.values():
            if robot.isOffline:
                continue

            positioner = self.fps[robot.id]
            robot.setAlphaBeta(positioner.alpha, positioner.beta)
            robot.setDestinationAlphaBeta(positioner.alpha, positioner.beta)

            if offsets.loc[robot.id].transformation_valid == 0:
                continue

            new = offsets.loc[robot.id, ["alpha_new", "beta_new"]]
            dist = offsets.loc[robot.id, ["xwok_distance", "ywok_distance"]] * 1000.0
            if numpy.hypot(dist.xwok_distance, dist.ywok_distance) > target_distance:
                robot.setDestinationAlphaBeta(new.alpha_new, new.beta_new)
            else:
                robot.isOffline = True

        # Check for collisions. If robots are collided just leave them there.
        collided = grid.getCollidedRobotList()
        n_coll = len(collided)
        if n_coll > 0:
            for pid in collided:
                positioner = self.fps[pid]
                alpha = positioner.alpha
                beta = positioner.beta
                grid.robotDict[pid].setAlphaBeta(alpha, beta)
                grid.robotDict[pid].setDestinationAlphaBeta(alpha, beta)

        # Generate trajectories.
        (to_destination, _, did_fail, deadlocks) = await get_path_pair_in_executor(
            grid,
            path_generation_mode="greedy",
            ignore_did_fail=True,
            stop_if_deadlock=True,
            ignore_initial_collisions=True,
        )
        if did_fail:
            log.warning(
                f"Found {len(deadlocks)} deadlocks but applying correction anyway."
            )

        self.log("Sending correction trajectory.")
        try:
            await self.fps.send_trajectory(to_destination, command=self.command)
        except TrajectoryError as err:
            raise FVCError(f"Failed executing the correction trajectory: {err}")

        self.correction_applied = True
        self.log("Correction applied.")

    async def write_summary_F(
        self,
        path: str | pathlib.Path | None = None,
        plot: bool = True,
        extra_headers: dict = {},
    ):
        """Updates data with the last measured positions and write confSummaryF."""

        if self.configuration is None:
            raise FVCError("write_summary_F can only be called with a configuration.")

        if self.fibre_data is None:
            raise FVCError("No fibre data.")

        self.log("Updating coordinates.", level=logging.DEBUG)

        fdata = (
            self.fibre_data.reset_index()
            .set_index(["positioner_id", "fibre_type"])
            .copy()
        )

        idx = pandas.IndexSlice
        cols = ["hole_id", "xwok_measured", "ywok_measured"]
        measured = fdata.loc[idx[:, "Metrology"], cols].dropna()

        for (pid, ftype), row in measured.iterrows():
            (alpha, beta), _ = wok_to_positioner(
                row.hole_id,
                self.fps.observatory,
                "Metrology",
                row.xwok_measured,
                row.ywok_measured,
            )
            if not numpy.isnan(alpha):
                fdata.loc[idx[pid, :], ["alpha", "beta"]] = (alpha, beta)

        configuration_copy = self.configuration.copy()
        configuration_copy.assignment_data.fibre_table = fdata.copy()

        for pid in measured.index.get_level_values(0).tolist():
            alpha, beta = fdata.loc[(pid, "Metrology"), ["alpha", "beta"]]
            for ftype in ["APOGEE", "BOSS", "Metrology"]:
                configuration_copy.assignment_data.positioner_to_icrs(
                    pid,
                    ftype,
                    alpha,
                    beta,
                    update=True,
                )

        if self.proc_hdu and "IPA" in self.proc_hdu.header:
            rotator_angle = round(self.proc_hdu.header["IPA"], 2)
        else:
            rotator_angle = -999.0

        headers = {
            "rotator_angle": rotator_angle,
            "fvc_centroid_method": self.centroid_method or "?",
            "fvc_rms": self.fitrms,
            "fvc_90_perc": self.perc_90,
            "fvc_percent_reached": self.fvc_percent_reached,
            "fvc_image_path": self.proc_image_path if self.proc_image_path else "",
        }
        headers.update(extra_headers)

        await configuration_copy.write_summary(
            path=path,
            flavour="F",
            headers=headers,
            overwrite=True,
        )

        # Plot analysis of FVC loop.
        if plot and self.proc_image_path:
            if self.configuration.assignment_data.boresight is None:
                self.log(
                    "Configuration does not have boresight set. "
                    "Cannot produce FVC plots.",
                    level=logging.WARNING,
                )
            else:
                self.log("Creating FVC plots", level=logging.DEBUG)

                outpath = str(self.proc_image_path).replace(".fits", "_distances.pdf")

                plot_fvc_distances(
                    self.configuration,
                    configuration_copy.assignment_data.fibre_table,
                    path=outpath,
                )


async def reprocess_configuration(
    configuration_id: int,
    path: pathlib.Path | str | None = None,
    centroid_method: str | None = None,
    use_suffix: bool = True,
):
    """Reprocesses the FVC image from a configuration with a different centroid method.

    Outputs a new ``confSummaryF`` file.

    Parameters
    ----------
    configuration_id
        The configuration ID for which to reprocess data. Must have an existing
        ``confSummaryF`` file in ``$SDSSCORE_DIR``.
    path
        The path where to write the new ``confSummaryF`` file. If `None`, defaults
        to ``$SDSSCORE_DIR``.
    centroid_method
        The centroid method to use, one of ``"nudge"``, ``"sep"``, ``"winpos"``,
        or ``"simple"``.
    use_suffix
        If `True`, the new ``confSummaryF`` path file will have a suffix
        including the centroid mode used.

    Returns
    -------
    path
        The path to the new ``confSummaryF`` file.

    """

    site = config["observatory"]
    confSummaryF_path = Configuration._get_summary_file_path(
        configuration_id,
        site,
        "F",
    )

    if not os.path.exists(confSummaryF_path):
        raise FileNotFoundError(
            f"Cannot find a confSummaryF file for {configuration_id}."
        )

    header, _ = read_confSummary(confSummaryF_path)

    design = Design(
        header["design_id"],
        epoch=header["epoch"],
        scale=header["focal_scale"],
    )

    fps = FPS.get_instance()
    assert not fps.can, "This function cannot be called on a running FPS instance."

    configuration = design.configuration
    configuration.configuration_id = configuration_id

    fvc = FVC(site)

    proc_fimg = header["fvc_image_path"]
    fimg = proc_fimg.replace("proc-", "")

    posangles = fits.getdata(proc_fimg, "POSANGLES")
    fiber_data = pandas.DataFrame(fits.getdata(proc_fimg, "FIBERDATA"))

    positioner_coords = {}
    for row in posangles:
        positioner_coords[row["positionerID"]] = (row["alphaReport"], row["betaReport"])

    fvc.process_fvc_image(
        fimg,
        positioner_coords,
        configuration=configuration,
        fibre_data=fiber_data,
        centroid_method=centroid_method,
        plot=False,
    )

    if path is None:
        path = confSummaryF_path

    path = str(path)

    if use_suffix:
        path = path.replace(".par", f"_{fvc.centroid_method}.par")

    await fvc.write_summary_F(
        path=path,
        plot=False,
        extra_headers={
            "MJD": header["MJD"],
            "obstime": header["obstime"],
            "temperature": header["temperature"],
        },
    )
